#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import ast
import argparse
import base64
import copy
import getpass
import logging
import math
import os
import pickle
import psutil
import re
import sys
import socket
import tempfile
import threading
import time
import traceback
import types
import urllib
from zmq.log.handlers import PUBHandler

import urllib.parse
import urllib.request
from collections.abc import Sequence, Mapping, Set, KeysView
from collections import defaultdict
from html.parser import HTMLParser
from io import FileIO, StringIO, BytesIO
from typing import Optional, List, Dict

import fasteners
import yaml
from tqdm import tqdm as ProgressBar

__all__ = ['get_output']

COLOR_CODE = {
    'ENDC': 0,  # RESET COLOR
    'BOLD': 1,
    'UNDERLINE': 4,
    'BLINK': 5,
    'INVERT': 7,
    'CONCEALD': 8,
    'STRIKE': 9,
    'GREY30': 90,
    'GREY40': 2,
    'GREY65': 37,
    'GREY70': 97,
    'GREY20_BG': 40,
    'GREY33_BG': 100,
    'GREY80_BG': 47,
    'GREY93_BG': 107,
    'DARK_RED': 31,
    'RED': 91,
    'RED_BG': 41,
    'LIGHT_RED_BG': 101,
    'DARK_YELLOW': 33,
    'YELLOW': 93,
    'YELLOW_BG': 43,
    'LIGHT_YELLOW_BG': 103,
    'DARK_BLUE': 34,
    'BLUE': 94,
    'BLUE_BG': 44,
    'LIGHT_BLUE_BG': 104,
    'DARK_MAGENTA': 35,
    'PURPLE': 95,
    'MAGENTA_BG': 45,
    'LIGHT_PURPLE_BG': 105,
    'DARK_CYAN': 36,
    'AUQA': 96,
    'CYAN_BG': 46,
    'LIGHT_AUQA_BG': 106,
    'DARK_GREEN': 32,
    'GREEN': 92,
    'GREEN_BG': 42,
    'LIGHT_GREEN_BG': 102,
    'BLACK': 30,
}


def colorstr(astr: str, color: Optional[str] = None) -> str:
    color_code = 0 if color is None else COLOR_CODE[color]
    if sys.platform == 'win32':
        return astr
    else:
        return f'\033[{color_code}m{astr}\033[0m'


def emphasize(msg: str, color: Optional[str] = None):
    level_color = 0 if color is None else COLOR_CODE[color]
    # display text within `` and `` in green
    if sys.platform == 'win32':
        return str(msg).replace('``', '')
    elif level_color == 0:
        return re.sub(r'``([^`]*)``', '\033[32m\\1\033[0m', str(msg))
    else:
        return re.sub(r'``([^`]*)``',
                      f'\033[0m\033[32m\\1\033[0m\033[{level_color}m', str(msg))


class ColoredFormatter(logging.Formatter):
    ''' A logging formatter that uses color to differntiate logging messages
    and emphasize texts. Texts that would be empahsized are quoted with
    double backslashes (`` ``).
    '''

    def __init__(self, msg: str):
        logging.Formatter.__init__(self, msg)
        #
        # color for different logging levels. The current terminal color
        # is used for INFO
        self.LEVEL_COLOR = {
            'DEBUG': 'DARK_CYAN',
            'WARNING': 'PURPLE',
            'ERROR': 'RED',
            'CRITICAL': 'RED_BG',
        }

    def format(self, record):
        level_name = record.levelname
        if level_name in self.LEVEL_COLOR:
            level_color = self.LEVEL_COLOR[level_name]
            record.color_levelname = colorstr(level_name, level_color)
            record.color_name = colorstr(record.name, level_color)
            record.color_msg = colorstr(
                emphasize(record.msg, level_color), level_color)
        else:
            # for INFO, use default color
            record.color_levelname = record.levelname
            record.color_msg = emphasize(record.msg)
        # for testing certain error message
        # if 'runtime' in record.color_msg:
        #     print(get_traceback())
        return logging.Formatter.format(self, record)


def short_repr(obj, noneAsNA=False):
    '''Return a short representation of obj for clarity.'''
    if obj is None:
        return 'unspecified' if noneAsNA else 'None'
    elif isinstance(obj, str) and len(obj) > 80:
        return '{}...{}'.format(obj[:60].replace('\n', '\\n'),
                                obj[-20:].replace('\n', '\\n'))
    elif isinstance(obj, (str, int, float, bool)):
        return repr(obj)
    elif hasattr(obj, '__short_repr__'):
        return obj.__short_repr__()
    elif isinstance(obj, Sequence):  # should be a list or tuple
        if len(obj) == 0:
            return '[]'
        elif len(obj) == 1:
            return f'{short_repr(obj[0])}'
        elif len(obj) == 2:
            return f'{short_repr(obj[0])}, {short_repr(obj[1])}'
        else:
            return f'{short_repr(obj[0])}, {short_repr(obj[1])}, ... ({len(obj)} items)'
    elif isinstance(obj, dict):
        if not obj:
            return ''
        elif len(obj) == 1:
            first_key = list(obj.keys())[0]
            return f'{short_repr(first_key)!r}:{short_repr(obj[first_key])!r}'
        else:
            first_key = list(obj.keys())[0]
            return f'{short_repr(first_key)}:{short_repr(obj[first_key])}, ... ({len(obj)} items)'
    elif isinstance(obj, KeysView):
        if not obj:
            return ''
        elif len(obj) == 1:
            return short_repr(next(iter(obj)))
        else:
            return f'{short_repr(next(iter(obj)))}, ... ({len(obj)} items)'
    # elif hasattr(obj, 'target_name'):
    #    return obj.target_name()
    else:
        ret = str(obj)
        if len(ret) > 40:
            return f'{repr(obj)[:35]}...'
        else:
            return ret


#
# SoS Workflow dictionary
#


class WorkflowDict(object):
    """A dictionary object that keeps all SoS workflow objects.

    IMPORTANT:

    Python does not allow the passing of a derived class of dict as globals
    to eval or exec. Doing so will result in strange behavior such as __builtins__
    not found. We then have to embed a real dictionary in WorkflowDict instead of
    deriving a dict from it.
    """

    def __init__(self, *args, **kwargs):
        self._dict = dict(*args, **kwargs)

    def dict(self):
        return self._dict

    def clear(self):
        self._dict.clear()

    def set(self, key, value):
        '''A short cut to set value to key without triggering any logging
        or warning message.'''
        if hasattr(value, 'labels'):
            if 'VARIABLE' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                    'SOS_DEBUG']:
                env.log_to_file(
                    'VARIABLE',
                    f"Set {key} to {short_repr(value)} with labels {short_repr(value.labels)}"
                )
        else:
            if 'VARIABLE' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                    'SOS_DEBUG']:
                env.log_to_file(
                    'VARIABLE',
                    f"Set {key} to {short_repr(value)} of type {value.__class__.__name__}"
                )
        self._dict[key] = value
        # if self._change_all_cap_vars is not None and key.isupper():
        #    self._check_readonly(key, value)

    def quick_update(self, obj):
        '''Update without sanity check etc. For fast internal update'''
        self._dict.update(obj)

    def update(self, obj):
        '''Redefine update to trigger logging message'''
        self._dict.update(obj)
        for k, v in obj.items():
            # if k.isupper():
            #    self._check_readonly(k, v)
            if env.verbosity > 2:
                self._log(k, v)

    def __contains__(self, key):
        return key in self._dict

    def __getattr__(self, attr):
        # for attributes that cannot be found, default to dictionary attribute
        # (e.g. keys, pop, get...)
        return getattr(self._dict, attr)

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        '''Set value to key, trigger logging and warning messages if needed'''
        if env.verbosity > 2:
            self._log(key, value)
        if env.config['run_mode'] == 'prepare':
            self._warn(key, value)
        if key in ('input', 'output', 'depends', '_input', '_output',
                   '_depends', '_runtime'):
            raise ValueError(f'Variable {key} can only be set by SoS')
        self.set(key, value)

    def _log(self, key, value):
        if 'VARIABLE' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                'SOS_DEBUG']:
            env.log_to_file('VARIABLE',
                            f'Set ``{key}`` = ``{short_repr(value)}``')

    def _warn(self, key, value):
        if key.startswith('_') and not key.startswith('__') and key not in (
                '_input', '_output', '_step', '_index', '_depends', '_runtime'):
            env.logger.warning(
                f'{key}: Variables with leading underscore is reserved for SoS temporary variables.'
            )

    def clone_selected_vars(self, selected=None):
        return {
            x: copy.deepcopy(y)
            for x, y in self._dict.items()
            if (not selected or x in selected) and pickleable(y, x)
        }


#
# Runtime environment
#


class RuntimeEnvironments(object):
    '''A singleton object that provides runtime environment for SoS.
    Atributes of this object include:

    logger:
        a logging object

    verbosity:
        a verbosity level object that sets the verbosity level of the logger

    logfile:
        name of logfile for the logger. default to no logfile.

    '''
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(RuntimeEnvironments, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        # sockets that will be initialized by the controller code
        self.zmq_context = None
        self.master_push_socket = None
        self.master_request_socket = None

        self._sub_idx = 0
        self._sub_envs = [{}]

        # this function is used by tests to reset environments
        # after finishing an test
        self.reset()

    def request_new(self):
        old_idx = self._sub_idx
        # if we can find an old empty env, use it
        for idx, subenv in enumerate(self._sub_envs):
            if idx != old_idx and not subenv:
                self.switch(idx)
                return idx, old_idx
        # otherwise allocate a new one
        self.switch(len(self._sub_envs))
        return self._sub_idx, old_idx

    def restore_to_old(self, new_idx, old_idx):
        self.switch(old_idx)
        env._sub_envs[new_idx].clear()

    def switch(self, idx):
        # save old env
        if idx == self._sub_idx:
            return
        self._sub_envs[self._sub_idx]['sos_dict'] = self.sos_dict
        self._sub_envs[self._sub_idx]['config'] = copy.deepcopy(env.config)
        self._sub_envs[self._sub_idx]['socket'] = env.__socket__ if hasattr(
            env, '__socket__') else None
        if len(self._sub_envs) <= idx:
            self._sub_envs.append({
                'sos_dict': WorkflowDict(),
                'config': copy.deepcopy(env.config),
                'socket': env.__socket__ if hasattr(env, '__socket__') else None
            })
        if not self._sub_envs[idx]:
            self._sub_envs[idx] = {
                'sos_dict': WorkflowDict(),
                'config': copy.deepcopy(env.config),
                'socket': env.__socket__ if hasattr(env, '__socket__') else None
            }
        self.sos_dict = self._sub_envs[idx]['sos_dict']
        env.config = self._sub_envs[idx]['config']
        env.__socket__ = self._sub_envs[idx]['socket']
        self._sub_idx = idx
        # env.logger.error(f"{os.getpid()} switch to {idx} {env.sos_dict.get('num', 'unknown')}")

    _exec_dir = None
    _temp_dir = os.path.join(tempfile.gettempdir(), getpass.getuser(), '.sos')

    def log_to_file(self, topic, msg=''):
        # if only one parameter is given, assuming ALL topic and topic is the message
        if not msg:
            self.logger.debug(topic)
            return
        if topic not in self.config['SOS_DEBUG'] and 'ALL' not in self.config[
                'SOS_DEBUG']:
            return
        self.logger.debug(topic + ' - ' + str(msg))

    def reset(self):
        # logger
        self._logger = None
        self._verbosity: int = 2
        self._logging_socket = None
        self._set_logger()

        #
        # run mode, this mode controls how SoS actions behave
        #
        self.config = defaultdict(str)
        self.config.update({
            'config_file': None,
            'output_dag': None,
            'output_report': None,
            'wait_for_task': None,
            'default_queue': '',
            'worker_procs': ['2'],
            'max_running_jobs': None,
            'sig_mode': 'default',
            'run_mode': 'run',
            'verbosity': 1,
            # determined later
            'master_id': '',
            'SOS_DEBUG': set(),
        })
        if 'SOS_DEBUG' in os.environ:
            self.config['SOS_DEBUG'] = set([
                x for x in os.environ['SOS_DEBUG'].split(',')
                if '.' not in x and x != '-'
            ])
        #
        # global dictionaries used by SoS during the
        # execution of SoS workflows
        self.sos_dict = WorkflowDict()
        # parameters of the workflow, which will be handled differently
        self.parameter_vars = set()
        #
        # this directory will be used by a lot of processes
        self.exec_dir = os.getcwd()

        os.makedirs(
            os.path.join(os.path.expanduser('~'), '.sos', 'tasks'),
            exist_ok=True)

    #
    # attribute logger
    #

    def _set_logger(self, unused=None):
        # create a logger, we current use the regular logger but we should
        # switch to multiprocessing.get_logger if we notice trouble in, for example,
        # logging from multiple processes.
        self._logger = logging.getLogger()
        # clear previous handler
        while self._logger.hasHandlers():
            self._logger.removeHandler(self._logger.handlers[0])
        levels = {
            0: logging.ERROR,
            1: logging.WARNING,
            2: logging.INFO,
            3: logging.DEBUG,
            4: logging.DEBUG,
            None: logging.INFO
        }
        # stop setting root logger always to DEBUG. If SOS_DEBUG is defined as PROFILE, we profile
        # in non-debug mode. Otherwise debugging leval will be set to DEBUG
        debug_mode = 'SOS_DEBUG' in os.environ and os.environ[
            'SOS_DEBUG'] and os.environ['SOS_DEBUG'] != 'PROFILE'
        self._logger.setLevel(
            logging.DEBUG if debug_mode else levels[self._verbosity])

        if self._logging_socket:
            socket_handler = PUBHandler(self._logging_socket)
            # debug informaiton and time is always written to the log file
            socket_handler.setLevel(levels[self._verbosity])

            # ch.setFormatter(logging.Formatter(
            #    '%(asctime)s: %(levelname)s: %(message)s'))
            self._logger.addHandler(socket_handler)
            # also log to file for debugging purposes
            # ch = logging.FileHandler(os.path.join(os.path.expanduser('~'), 'sos_debug.log'), mode='a')
            # debug informaiton and time is always written to the log file
            # ch.setLevel(logging.TRACE)
            # ch.setFormatter(logging.Formatter(
            #    '%(asctime)s: %(levelname)s: %(message)s'))
            # self._logger.addHandler(ch)
        else:
            # output to standard output
            cout = logging.StreamHandler()
            cout.setLevel(levels[self._verbosity])
            cout.setFormatter(
                ColoredFormatter('%(color_levelname)s: %(color_msg)s'))
            self._logger.addHandler(cout)

        if debug_mode:
            logfile_info = [
                x for x in os.environ['SOS_DEBUG'].split(',')
                if '.' in x or x == '-'
            ]
            if logfile_info:
                if logfile_info[0] == '-':
                    logfile = logging.StreamHandler()
                    formatter = ColoredFormatter(
                        '%(color_levelname)s: %(color_msg)s')
                else:
                    logfile = logging.FileHandler(
                        os.path.expanduser(logfile_info[0]), mode='a')
                    formatter = logging.Formatter(
                        '%(asctime)s - %(levelname)s - %(message)s')
            else:
                logfile = logging.FileHandler(
                    os.path.join(
                        os.path.expanduser('~'), '.sos', 'sos_debug.log'),
                    mode='a')
                formatter = logging.Formatter(
                    '%(asctime)s - %(levelname)s - %(message)s')
            logfile.setFormatter(formatter)
            logfile.setLevel(logging.DEBUG)
            self._logger.addHandler(logfile)

    #
    # attribute exec_dir

    def _set_exec_dir(self, edir):
        if not os.path.isdir(edir):
            raise RuntimeError(f'Exec dir {edir} does not exist.')
        os.makedirs(os.path.join(edir, '.sos'), exist_ok=True)
        self._exec_dir = edir

    def _get_exec_dir(self):
        if self._exec_dir is None:
            raise RuntimeError('Exec dir is not set')
        os.makedirs(os.path.join(self._exec_dir, '.sos'), exist_ok=True)
        return self._exec_dir

    exec_dir = property(_get_exec_dir, _set_exec_dir)

    # attribute temp_dir
    def _get_temp_dir(self):
        if not os.path.isdir(self._temp_dir):
            os.makedirs(self._temp_dir, exist_ok=True)
        return self._temp_dir

    temp_dir = property(_get_temp_dir)

    #
    # attribute logger
    #
    @property
    def logger(self):
        return self._logger

    #
    # attribute verbosity
    #

    def _set_verbosity(self, v):
        if v in [0, 1, 2, 3, 4]:
            self._verbosity = v
            # reset logger to appropriate logging level
            self._set_logger()

    #
    verbosity = property(lambda self: self._verbosity, _set_verbosity)

    #
    # attribute logfile
    #

    def set_socket_logger(self, socket):
        self._logging_socket = socket
        # reset logger to include log file
        self._set_logger()


# set up environment variable and a default logger
env = RuntimeEnvironments()


def get_logger():
    return env.logger


#
# String formatting
#
class _DeHTMLParser(HTMLParser):
    '''This parser analyzes input text, removes HTML tags such as
    <p>, <br>, <ul>, <li> etc and returns properly formatted texts.
    '''

    def __init__(self):
        HTMLParser.__init__(self)
        self.__text = []

    def handle_data(self, data):
        text = data.strip()
        if len(text) > 0:
            text = re.sub('[ \t\r\n]+', ' ', text)
            self.__text.append(text + ' ')

    def handle_starttag(self, tag, attrs):
        if tag == 'p':
            self.__text.append('\n\n\n\n')
        elif tag == 'br':
            self.__text.append('\n\n')
        elif tag == 'ul':
            self.__text.append('')
        elif tag == 'li':
            self.__text.append('\n\n  * ')

    def handle_endtag(self, tag):
        if tag == 'ul':
            self.__text.append('\n\n')
        if tag == 'li':
            self.__text.append('\n\n')

    def handle_startendtag(self, tag, attrs):
        if tag == 'br':
            self.__text.append('\n\n')

    def text(self):
        return ''.join(self.__text).strip()


def dehtml(text):
    '''Remove HTML tag in input text and format the texts
    accordingly. '''
    try:
        parser = _DeHTMLParser()
        parser.feed(text)
        parser.close()
        return parser.text()
    except Exception as e:
        env.logger.warning(f'Failed to dehtml text: {e}')
        return text


# exception classes


class Error(Exception):
    '''Base class for SoS_ScriptParser exceptions.'''

    def _get_message(self) -> str:
        '''Getter for 'message'; needed only to override deprecation in
        BaseException.'''
        return self.__message

    def _set_message(self, value: str) -> None:
        '''Setter for 'message'; needed only to override deprecation in
        BaseException.'''
        self.__message = value

    # BaseException.message has been deprecated since Python 2.6.  To prevent
    # DeprecationWarning from popping up over this pre-existing attribute, use
    # a new property that takes lookup precedence.
    message = property(_get_message, _set_message)

    def __init__(self, msg: str = '') -> None:
        self.message: str = msg
        Exception.__init__(self, msg)

    def __repr__(self) -> str:
        return self.message

    __str__ = __repr__


class StopInputGroup(Error):
    '''Abort a step and continue'''

    def __init__(self, msg, keep_output=False):
        self.keep_output = keep_output
        Error.__init__(self, msg)


class TerminateExecution(Error):
    '''Abort a step and continue'''

    def __init__(self, msg):
        Error.__init__(self, msg)


class ProcessKilled(Error):
    '''Process is killed'''

    def __init__(self, msg=''):
        Error.__init__(self, msg)


def get_traceback():
    output = StringIO()
    exc_type, exc_value, exc_traceback = sys.exc_info()
    # print "*** print_tb:"
    traceback.print_tb(exc_traceback, limit=1, file=output)
    # print "*** print_exception:"
    try:
        traceback.print_exception(
            exc_type, exc_value, exc_traceback, limit=5, file=output)
    except Exception:
        # the above function call can fail under Python 3.4 for some
        # exception but we do not really care if that happens
        pass
    result = output.getvalue()
    output.close()
    return result
    # print "*** print_exc:"
    # traceback.print_exc()
    # print "*** format_exc, first and last line:"
    # formatted_lines = traceback.format_exc().splitlines()
    # print formatted_lines[0]
    # print formatted_lines[-1]
    # print "*** format_exception:"
    # print repr(traceback.format_exception(exc_type, exc_value,
    #                                      exc_traceback))
    # print "*** extract_tb:"
    # print repr(traceback.extract_tb(exc_traceback))
    # print "*** format_tb:"
    # print repr(traceback.format_tb(exc_traceback))
    # print "*** tb_lineno:", exc_traceback.tb_lineno


def pickleable(obj, name):
    if isinstance(obj, (str, bool, int, float, complex, bytes)):
        return True
    if isinstance(obj, (types.ModuleType, WorkflowDict)):
        return False
    if callable(obj):
        return False
    try:
        pickle.dumps(obj)
        return True
    except Exception as e:
        env.logger.debug(
            f'Object {name} with value {short_repr(obj)} is not passed because it is not pickleable: {e}'
        )
        return False


class ProgressFileObj(FileIO):
    '''A wrapper of a file object that update a progress bar
    during file read.
    '''

    def __init__(self, prog, *args, **kwargs):
        FileIO.__init__(self, *args, **kwargs)
        self.prog = prog

    def read(self, n, *args):
        self.prog.update(n)
        return FileIO.read(self, n, *args)


def stable_repr(obj):
    if isinstance(obj, str):
        return repr(obj)
    if hasattr(obj, '__stable_repr__'):
        return obj.__stable_repr__()
    if isinstance(obj, Mapping):
        items = [stable_repr(k) + ':' + stable_repr(obj[k]) for k in obj.keys()]
        return '{' + ', '.join(sorted(items)) + '}'
    if isinstance(obj, Set):
        items = [stable_repr(x) for x in obj]
        return '{' + ', '.join(sorted(items)) + '}'
    if isinstance(obj, Sequence):
        return '[' + ', '.join(stable_repr(k) for k in obj) + ']'
    return repr(obj)


#
# A utility function that returns output of a command


def get_output(cmd, show_command=False, prompt='$ ', **kwargs):
    import subprocess
    try:
        if 'shell' not in kwargs:
            kwargs['shell'] = True
        output = subprocess.check_output(
            cmd, stderr=subprocess.DEVNULL, **kwargs).decode()
    except subprocess.CalledProcessError as e:
        if e.output.decode():
            env.logger.error(e.output.decode())
        raise RuntimeError(e)
    if show_command:
        return f'{prompt}{cmd}\n{output}'
    else:
        return output


#
# search a path and locate script and other files
#


def locate_script(filename, start=''):
    #
    attemp = os.path.abspath(os.path.expanduser(filename))
    if os.path.isfile(attemp):
        return ('', attemp)
    #
    token = urllib.parse.urlparse(filename)
    # if no scheme or netloc, the URL is not acceptable
    if all([
            getattr(token, qualifying_attr)
            for qualifying_attr in ('scheme', 'netloc')
    ]):
        try:
            local_filename, _ = urllib.request.urlretrieve(filename)
            with open(local_filename) as script:
                content = script.read()
            #
            return (content, filename)
        except Exception as e:
            env.logger.error(str(e))
            raise ValueError(f'Failed to open {filename}')
    #
    # a search path
    pathes = [start]
    sos_config_file = os.path.join(
        os.path.expanduser('~'), '.sos', 'config.yml')
    if os.path.isfile(sos_config_file):
        try:
            with open(sos_config_file) as config:
                cfg = yaml.safe_load(config)
        except Exception:
            raise RuntimeError(
                f'Failed to parse global sos config file {sos_config_file}, is it in JSON format?'
            )
        #
        pathes.extend(cfg.get('sos_path', []))
    #
    for path in pathes:
        if not path:
            continue
        attemp = os.path.join(
            os.path.expanduser(path), os.path.expanduser(filename))
        if os.path.isfile(attemp):
            return ('', attemp)
        # is it an URL?
        token = urllib.parse.urlparse(path)
        # if no scheme or netloc, the URL is not acceptable
        if all([
                getattr(token, qualifying_attr)
                for qualifying_attr in ('scheme', 'netloc')
        ]):
            url = path + ('' if path.endswith('/') else '/') + filename
            try:
                local_filename, _ = urllib.request.urlretrieve(url)
                with open(local_filename) as script:
                    content = script.read()
                return content, url
            except Exception:
                pass
    #
    raise ValueError(f'Failed to locate {filename}')


def valid_expr_till(text):
    pos = len(text)
    while pos > 0:
        try:
            if not text[:pos].lstrip():
                return 0
            # so we trying to find out the valid expression before : and !
            # but somehow a : r is a valid expression until we put it inside ().
            ast.parse('(' + text[:pos].lstrip() + ')')
            if pos == len(text) or text[pos] == '!' or text[pos] == ':':
                return pos
            else:
                return 0
        except Exception:
            pos -= 1
    return 0


def check_last_piece(text):
    pos = 0
    while True:
        spos = text.find('}', pos)
        if spos == -1:
            return True
        elif spos == len(text) - 1 or text[spos + 1] != '}':
            raise SyntaxError("f-string: single '}' is not allowed")
        elif spos == len(text) - 2:
            # }} as the last
            return True
        else:
            pos = spos + 2


def split_fstring(text):
    # now that we have the correct sigil
    # first, we need to replace all { as {{ and } as }}
    pieces = []
    pos = 0
    while True:
        dpos = text.find('{{', pos)
        spos = text.find('{', pos)
        if spos == -1:
            # no more {
            check_last_piece(text)
            pieces.append(text)
            break
        if spos == dpos:  # no { before {{
            pos = dpos + 2
            continue
        # now we have a valid spos
        pieces.append(text[:spos])
        # skip '{'
        rhs_pieces = text[spos + 1:].split('}')
        if len(rhs_pieces) == 1:
            raise SyntaxError(
                f'Invalid f-string {text}: missing right sigil at {text[pos:pos+20]}'
            )
        # rhs = 'whatever }' :r} something else {}
        #
        # we need to include } in expression
        for n in range(1, len(rhs_pieces) + 1):
            if valid_expr_till('}'.join(rhs_pieces[:n])) > 0:
                pieces.append('}'.join(rhs_pieces[:n]))
                text = '}'.join(rhs_pieces[n:])
                break
            # the last one, still not valid
            if n == len(rhs_pieces):
                raise SyntaxError(
                    f'Invalid f-string "{text}": invalid or empty expression')
    return pieces


def as_fstring(text):
    """expansion with python f-string, usually ok, but with the case
       of ' inside expressions, adding repr will add backslash to it
       and cause trouble.
    """
    for quote in ('"""', "'''"):
        # although script-format always ends with \n, direct use of this function might
        # have string ending with " or '
        if quote not in text and not text.endswith(quote[0]):
            return 'fr' + quote + text + quote

    # now, we need to look into the structure of f-string
    pieces = split_fstring(text)
    # if all expressions do not have single quote, we can
    # safely repr the entire string
    if not any("'" in piece for piece in pieces[1::2]):
        return 'f' + repr(text)
    #
    # unfortunately, this thing has both single and double triple quotes
    # because we cannot use backslash inside expressions in f-string, we
    # have to use format string now.
    args = []
    for idx in range(len(pieces))[1::2]:
        pos = valid_expr_till(pieces[idx])
        if pos == 0:
            raise SyntaxError(f'invalid expression in {pieces[idx]}')
        args.append(pieces[idx][:pos])
        pieces[idx] = '{' + str(idx // 2) + pieces[idx][pos:] + '}'
    return repr(''.join(pieces)) + '.format(' + ', '.join(args) + ')'


def natural_keys(text):
    '''
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    '''
    return [int(c) if c.isdigit() else c for c in re.split(r'(\d+)', text)]


def transcribe(text, cmd=None):
    if cmd is not None:
        text = '{}:\n{}'.format(cmd,
                                '    ' + text.replace('\n', '\n    ') + '\n')
    with fasteners.InterProcessLock(
            os.path.join(env.temp_dir, 'transcript.lck')):
        with open(os.path.join(env.exec_dir, '.sos', 'transcript.txt'),
                  'a') as trans:
            trans.write(text)


def dict_merge(dct, merge_dct):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for k, v in merge_dct.items():
        if (k in dct and isinstance(dct[k], dict) and isinstance(v, dict)):
            dict_merge(dct[k], v)
        else:
            dct[k] = v


# display file size in K, M, G etc automatically. Code copied from
# http://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
# for its compact size


def pretty_size(n,
                pow=0,
                b=1024,
                u='B',
                pre=[''] + [p + 'i' for p in 'KMGTPEZY']):
    pow, n = min(int(math.log(max(n * b**pow, 1), b)), len(pre) - 1), n * b**pow
    return "%%.%if %%s%%s" % abs(pow %
                                 (-pow - 1)) % (n / b**float(pow), pre[pow], u)


def expand_size(size):
    if isinstance(size, int):
        return size
    m = re.match(r'\s*([+-]?)([\.\d]*)\s*(\S+)\s*', size)
    if not m:
        raise ValueError(f'Invalid size specified: {size}')
    sign, num, unit = m.groups()
    sign = -1 if sign == '-' else 1
    if not unit:
        return sign * int(num)
    if not num:
        num = 1
    s = {x + 'I': 1024**(idx + 1) for idx, x in enumerate('KMGTPEZY')}
    s.update({x: 1000**(idx + 1) for idx, x in enumerate('KMGTPEZY')})
    unit = unit[:-1].upper() if unit[-1].upper().endswith('B') else unit.upper()
    if unit not in s:
        raise ValueError(f'Invalid size specified: {size}')
    return sign * int(float(num) * s[unit])


def find_symbolic_links(item):
    item = os.path.expanduser(item)
    if os.path.islink(item):
        if not os.path.exists(item):
            env.logger.warning(f'Non-existent symbolic link {item}')
        return {item: os.path.realpath(item)}
    elif os.path.isfile(item):
        return {}
    else:
        result = {}
        for x in os.listdir(item):
            result.update(find_symbolic_links(os.path.join(item, x)))
        return result


class ActivityNotifier(threading.Thread):

    def __init__(self, msg, delay=5):
        threading.Thread.__init__(self)
        self.msg = msg
        self.start_time = time.time()
        self.delay = delay
        self.event = threading.Event()
        self.start()

    def run(self):
        prog = None
        while True:
            self.event.wait(self.delay)
            if self.event.is_set():
                if prog:
                    prog.close()
                break
            if not prog:
                print(self.msg)
                prog = ProgressBar(
                    desc='', position=0, bar_format='{desc}', total=100000000)
            second_elapsed = time.time() - self.start_time
            prog.set_description('Elapsed time {}{}'.format(
                '' if second_elapsed < 86400 else
                f'{int(second_elapsed/86400)} day{"s" if second_elapsed > 172800 else ""} ',
                time.strftime('%H:%M:%S', time.gmtime(second_elapsed))))
            prog.update(1)

    def stop(self):
        self.event.set()


class DelayedAction:
    '''Call the passed function with param after a few seconds. It is most often
    used to display certain message only if an action takes a long time.

        action = delayedAction(env.logger.info, 'This might take a while', 5)
        some_action_that_might_take_a_while
        del action

    if the action finishes very quick, the message will not be displayed.
    '''

    def __init__(self, func, param, delay=5):
        self.timer = threading.Timer(delay, func, (param,))
        self.timer.start()

    def __del__(self):
        self.timer.cancel()


class ArgumentError(Error):
    """Raised when an invalid argument is passed."""

    def __init__(self, msg):
        Error.__init__(self, msg)
        self.args = (msg,)


def _parse_error(msg):
    '''This function will replace error() function in argparse module so that SoS
    can hijack errors raised from it.'''
    raise ArgumentError(msg)


def sos_get_param(key, defvalue):
    '''Parse command line arguments and set values to parameters section.
    NOTE: parmeters will not be handled if it is already defined in
    the environment. This makes the parameters variable.
    '''
    env.parameter_vars.add(key)
    # if no argument is provided
    if not env.config['workflow_args'] and not env.config['workflow_vars']:
        if isinstance(defvalue, type):
            raise ArgumentError(
                f'Argument {key} of type {defvalue.__name__} is required')
        else:
            return defvalue
    #
    if env.config['workflow_vars']:
        if key in env.config['workflow_vars']:
            return env.config['workflow_vars'][key]

    parser = argparse.ArgumentParser(allow_abbrev=False)
    # thre is a possibility that users specify --cut-off instead of --cut_off for parameter
    # cut_off. It owuld be nice to allow both.
    #
    # Argparse would produce cut_off for both definition of --cut-off and --cut_off, however
    # you can only use the matching input...
    from .targets import path, paths, sos_targets, file_target
    ret_type = None

    if isinstance(defvalue, type) or defvalue is None:
        if defvalue == bool:
            feature_parser = parser.add_mutually_exclusive_group(required=True)
            feature_parser.add_argument(
                f'--{key}', dest=key, action='store_true')
            feature_parser.add_argument(
                f'--no-{key}', dest=key, action='store_false')
            if '_' in key:
                feature_parser.add_argument(
                    f'--{key.replace("_", "-")}', dest=key, action='store_true')
                feature_parser.add_argument(
                    f'--no-{key.replace("_", "-")}',
                    dest=key,
                    action='store_false')
        else:
            if defvalue is None:
                defvalue = str
            if defvalue in (sos_targets, paths):
                ret_type = defvalue
            # if only a type is specified, it is a required document of required type
            if '_' in key:
                feature_parser = parser.add_mutually_exclusive_group(
                    required=True)
                feature_parser.add_argument(
                    f'--{key}',
                    dest=key,
                    type=str if hasattr(defvalue, '__iter__') and
                    defvalue not in (file_target, path) else defvalue,
                    help='',
                    nargs='+' if defvalue not in (str, file_target) and
                    hasattr(defvalue, '__iter__') else '?')
                feature_parser.add_argument(
                    f'--{key.replace("_", "-")}',
                    dest=key,
                    type=str if hasattr(defvalue, '__iter__') and
                    defvalue not in (file_target, path) else defvalue,
                    help='',
                    nargs='+' if defvalue not in (str, file_target) and
                    hasattr(defvalue, '__iter__') else '?')
            else:
                parser.add_argument(
                    f'--{key}',
                    dest=key,
                    type=str if hasattr(defvalue, '__iter__') and
                    defvalue not in (file_target, path) else defvalue,
                    help='',
                    required=True,
                    nargs='+' if defvalue not in (str, file_target, path) and
                    hasattr(defvalue, '__iter__') else '?')
    else:
        if isinstance(defvalue, bool):
            feature_parser = parser.add_mutually_exclusive_group(required=False)
            feature_parser.add_argument(
                f'--{key}', dest=key, action='store_true')
            feature_parser.add_argument(
                f'--no-{key}', dest=key, action='store_false')
            if '_' in key:
                feature_parser.add_argument(
                    f'--{key.replace("_", "-")}', dest=key, action='store_true')
                feature_parser.add_argument(
                    f'--no-{key.replace("_", "-")}',
                    dest=key,
                    action='store_false')
            feature_parser.set_defaults(**{key: defvalue})
        else:
            if isinstance(defvalue, (file_target, path)):
                deftype = type(defvalue)
            elif isinstance(defvalue, paths):
                deftype = path
                ret_type = type(defvalue)
            elif isinstance(defvalue, sos_targets):
                deftype = file_target
                ret_type = type(defvalue)
            elif isinstance(defvalue, str):
                deftype = str
            elif isinstance(defvalue, Sequence):
                if len(defvalue) > 0:
                    deftype = type(defvalue[0])
                else:
                    deftype = str
            else:
                deftype = type(defvalue)

            if '_' in key:
                feature_parser = parser.add_mutually_exclusive_group(
                    required=False)
                feature_parser.add_argument(
                    f'--{key}',
                    dest=key,
                    type=deftype,
                    nargs='*' if isinstance(defvalue, Sequence) and
                    not isinstance(defvalue, str) else '?',
                    default=defvalue)
                feature_parser.add_argument(
                    f'--{key.replace("_", "-")}',
                    dest=key,
                    type=deftype,
                    nargs='*' if isinstance(defvalue, Sequence) and
                    not isinstance(defvalue, str) else '?',
                    default=defvalue)
            else:
                parser.add_argument(
                    f'--{key}',
                    dest=key,
                    type=deftype,
                    nargs='*' if isinstance(defvalue, Sequence) and
                    not isinstance(defvalue, str) else '?',
                    default=defvalue)
    #
    parser.error = _parse_error
    parsed, _ = parser.parse_known_args(env.config['workflow_args'])
    return ret_type(vars(parsed)[key]) if ret_type else vars(parsed)[key]


# def is_locked(lockfile):
#    lock = fasteners.InterProcessLock(lockfile)
#    gotten = lock.acquire(blocking=False)
#    if gotten:
#        lock.release()
#        return False
#    else:
#        return True


class TimeoutInterProcessLock(fasteners.InterProcessLock):
    #
    # #871
    #
    # For some unknown reason, sos could hang after tasks fail to obtain a lock.
    # The problem should **NOT** be able to fix by simply removing the lock file
    # because the new lock file should still be marked as "locked" if the process
    # that clocks the lock is still valid. However, we have observed cases that
    # no process could be found to be locking the file, yet the lock appears to
    # be occupied, and yet removing the lock file **will not** automatically
    # unlock the waiting process, yet rerunning sos after removing the lock
    # file would work.
    #
    # This is really strange but we are **temporarily** replacing fasteners.InterProcessLock
    # with this TimeoutInterProcessLock which would try to remove the lock file after
    # timeout (default to 5) seconds, and try to obtain a lock after the removal of the
    # lock file.
    #
    def __init__(self, path, timeout=5, sleep_func=time.sleep, logger=None):
        super(TimeoutInterProcessLock, self).__init__(
            path, sleep_func=sleep_func, logger=logger)
        self.timeout = timeout

    def __enter__(self):
        start_time = time.time()
        msg = False
        while True:
            gotten = self.acquire(blocking=False)
            if gotten:
                return self
            self.sleep_func(0.01)
            if time.time() - start_time > self.timeout:
                if os.path.exists(self.path):
                    try:
                        os.remove(self.path)
                    except Exception:
                        pass
                if not msg:
                    env.logger.warning(
                        f'Failed to obtain lock {self.path} after {self.timeout} seconds, perhaps you will have to remove the lock file manually.'
                    )
                    msg = True


config_cache: dict = {}


def load_config_files(filename=None, default_config_files=True):
    # user-specified configuration file.
    if filename is None and 'config_file' in env.config:
        filename = env.config['config_file']

    if filename is not None and not os.path.isfile(
            os.path.expanduser(filename)):
        env.logger.warning(f'Ignoring missing configuration file {filename}')
        filename = None

    filemtime = None if filename is None else os.path.getmtime(
        os.path.expanduser(filename))

    from .targets import textMD5
    extra_cfg = None if 'extra_config' not in env.config else textMD5(
        str(env.config['extra_config']))

    if (filename, filemtime, extra_cfg) in config_cache:
        env.sos_dict.set('CONFIG',
                         config_cache[(filename, filemtime, extra_cfg)])
        return config_cache[(filename, filemtime, extra_cfg)]

    cfg = {}
    # site configuration file
    if default_config_files:
        sos_config_file = os.path.join(
            os.path.split(__file__)[0], 'site_config.yml')
        if os.path.isfile(sos_config_file):
            try:
                with open(sos_config_file) as config:
                    cfg = yaml.safe_load(config)
            except Exception as e:
                raise RuntimeError(
                    f'Failed to parse global sos hosts file {sos_config_file}, is it in YAML/JSON format? ({e})'
                )

        # global site file
        sos_config_file = os.path.join(os.path.expanduser('~'), '.sos', 'hosts.yml')
        if os.path.isfile(sos_config_file):
            try:
                with open(sos_config_file) as config:
                    gd = yaml.safe_load(config)
                dict_merge(cfg, gd)
            except Exception as e:
                print(get_traceback())
                env.logger.warning(
                    f'Failed to parse global sos hosts file {sos_config_file}: {e}')
        # global config file
        sos_config_file = os.path.join(
            os.path.expanduser('~'), '.sos', 'config.yml')
        if os.path.isfile(sos_config_file):
            try:
                with open(sos_config_file) as config:
                    gd = yaml.safe_load(config)
                dict_merge(cfg, gd)
            except Exception as e:
                env.logger.warning(
                    f'Failed to parse global sos config file {sos_config_file}: {e}'
                )

    if filename is not None:
        if not os.path.isfile(os.path.expanduser(filename)):
            raise RuntimeError(f'Config file {filename} not found')
        try:
            with open(os.path.expanduser(filename)) as config:
                gd = yaml.safe_load(config)
            dict_merge(cfg, gd)
        except Exception as e:
            env.logger.warning(f'Failed to parse config file {filename}: {e}')

    if 'user_name' not in cfg:
        cfg['user_name'] = getpass.getuser().lower()

    if 'extra_config' in env.config and isinstance(env.config['extra_config'],
                                                   dict):
        cfg.update(env.config['extra_config'])

    # handle keyword "based_on", which should fill the dictionary with others.
    def process_based_on(cfg, item):
        if 'based_on' in item:
            if not isinstance(item['based_on'],
                              (str, list)) or not item['based_on']:
                raise ValueError(
                    f'A string is expected for key based_on. {item["based_on"]} obtained'
                )

            referred_keys = [item['based_on']] if isinstance(
                item['based_on'], str) else item['based_on']
            item.pop('based_on')
            for rkey in referred_keys:
                # find item...
                val = cfg
                for key in rkey.split('.'):
                    if not isinstance(val, dict):
                        raise ValueError(f'Based on key {item} not found')
                    if key not in val:
                        raise ValueError(
                            f'Based on key {key} not found in config')
                    else:
                        val = val[key]
                #
                if not isinstance(val, dict):
                    raise ValueError('Based on item must be a dictionary')
                if 'based_on' in val:
                    val = process_based_on(cfg, val)
                # ok, we have got a dictionary, let us use it to replace item
                for k, v in val.items():
                    if k not in item:
                        item[k] = v
            return item
        else:
            for k, v in item.items():
                if isinstance(v, dict):
                    # v should be processed in place
                    process_based_on(cfg, v)
            return item

    #
    for v in cfg.values():
        if isinstance(v, dict):
            process_based_on(cfg, v)

    config_cache[(filename, filemtime, extra_cfg)] = cfg
    env.sos_dict.set('CONFIG', cfg)
    return cfg


def format_duration(time_diff_secs, short=True):
    secs = int(time_diff_secs)
    rec = [(secs // (3600 * 24), 'day'), (secs % (3600 * 24) // 3600, 'hr'),
           (secs % 3600 // 60, 'min'), (secs % 60, 'sec')]
    txt = [f'{x} {y}' for x, y in rec if x > 0]
    if short and len(txt) > 1:
        txt = [x for x in txt[:2] if not x.endswith('sec')]
    return ' '.join(txt) if txt else '0s'


def format_HHMMSS(v):
    if isinstance(v, int):
        h, m, s = v // 3600, v % 3600 // 60, v % 60
    elif isinstance(v, str):
        # the time can be spacified as age.
        try:
            return format_HHMMSS(expand_time(v))
        except Exception as e:
            raise ValueError(
                f'walltime should be specified as a integer with unit s (default), h, m, d or string in the format of HH:MM:SS. "{v}" specified ({e})'
            )
    else:
        raise ValueError(
            f'walltime should be specified as a integer with unit s (default), h, m, d or string in the format of HH:MM:SS. "{v}" of type {v.__class__.__name__} specified'
        )
    return f'{h:02d}:{m:02d}:{s:02d}'


def expand_time(v, default_unit='s') -> int:
    # expand walltime from '00:00:00' format to second
    if isinstance(v, str):
        try:
            sign = {'+': 1, '-': -1}[v[1]]
            v = v[1:]
        except Exception:
            # if there is no sign, assume +
            sign = 1

        # format HHMMSS
        if ':' in v:
            try:
                h, m, s = map(int, v.split(':'))
                return sign * (h * 60 * 60 + m * 60 + s)
            except Exception as e:
                raise ValueError(
                    f'Input of option walltime should be an integer with unit s (default), h, m, d or a string in the format of HH:MM:SS. {v} specified: {e}'
                )
        #
        try:
            unit = {'s': 1, 'm': 60, 'h': 3600, 'd': 3600 * 24}[v[-1]]
            v = v[:-1]
        except Exception:
            unit = {'s': 1, 'm': 60, 'h': 3600, 'd': 3600 * 24}[default_unit]
        #
        try:
            return int(sign * unit * float(v))
        except Exception:
            raise ValueError(
                f'Unacceptable time for parameter age, expecting [+/-] num [s|m|h|d] or HH:MM:SS (e.g. +5h): {v} provided'
            )
    elif isinstance(v, int):
        return v
    else:
        raise ValueError(
            f'Input of option walltime should be an integer with unit s (default), h, m, d or a string in the format of HH:MM:SS. {v} specified.'
        )


def tail_of_file(filename, n, ansi2html=False):
    """Reads a n lines from f with an offset of offset lines. """
    avg_line_length = 74
    to_read = n

    with open(filename) as f:
        while 1:
            try:
                f.seek(-(avg_line_length * to_read), 2)
            except IOError:
                # woops.  apparently file is smaller than what we want
                # to step back, go to the beginning instead
                f.seek(0)
            pos = f.tell()
            lines = f.read().splitlines()
            if len(lines) >= to_read or pos == 0:
                if ansi2html:
                    return convertAnsi2html('\n'.join(lines[-to_read:]))
                return '\n'.join(lines[-to_read:]) + '\n'
            avg_line_length *= 1.3


def sample_lines(lines, n):
    '''Draw a sample of n lines from filename, largely evenly.'''
    if len(lines) <= n:
        return ''.join(lines)
    else:
        m = len(lines)
        return ''.join([lines[x * m // n + m // (2 * n)] for x in range(n)])


def linecount_of_file(filename):
    f = open(filename, 'rb')
    lines = 0
    buf_size = 1024 * 1024
    read_f = f.raw.read

    buf = read_f(buf_size)
    while buf:
        lines += buf.count(b'\n')
        buf = read_f(buf_size)

    return lines


def isPrimitive(obj):
    # test if object is of primitive types (string, number, sequence etc
    # http://stackoverflow.com/questions/6391694/check-if-a-variables-type-is-primitive
    return not hasattr(obj, '__dict__')


def save_var(name, var):
    if isinstance(var, (bool, int, float, complex, str, bytes)) or var is None:
        return f'{name}={repr(var)}\n'
    if isinstance(var, (type, types.ModuleType, WorkflowDict, Exception)):
        return ''
    try:
        # for more complex type, we use pickle + base64
        return f'{name}:={base64.b64encode(pickle.dumps(var))}\n'
    except Exception:
        return ''


def load_var(line):
    from .targets import remote
    assert remote
    key, value = line.split('=', 1)
    if key.endswith(':'):
        return key[:-1], pickle.loads(base64.b64decode(eval(value.strip())))
    else:
        try:
            return key, eval(value.strip())
        except Exception:
            # use SoS_eval instead of eval because vars can contain sos objects such as R_library
            from .eval import SoS_eval
            return key, SoS_eval(value.strip())


def version_info(module: str):
    # return the version of Python module
    try:
        code = ("import %s; version=str(%s.__version__)" % (module, module))
        ns_g: Dict = {}
        ns_l: Dict = {}
        exec(compile(code, "<string>", "exec"), ns_g, ns_l)
        return ns_l["version"]
    except Exception:
        import pkg_resources
        try:
            return pkg_resources.require(module)[0].version
        except Exception:
            return 'na'


def loaded_modules(namespace=None):
    if not namespace:
        return []
    res = {}
    for value in namespace.values():
        if isinstance(value, types.ModuleType):
            res[value.__name__] = version_info(value.__name__)
    return [(x, y) for x, y in res.items() if y != 'na']


def convertAnsi2html(txt):
    # 94 is blue, debug
    # 32 is darkgreen, emphasize
    # 95 is purple, warning
    # 91 is red, error
    # 36 is cray, trace
    return txt.replace('\033[94m', '<font color="">'). \
        replace('\033[32m', '<font color="DarkGreen">'). \
        replace('\033[36m', '<font color="cyan">'). \
        replace('\033[95m', '<font color="purple">'). \
        replace('\033[91m', '<font color="red">'). \
        replace('\033[0m', '</font>'). \
        replace('\n', '<br>')


# log to file for debugging purpose only


def pexpect_run(cmd, shell=False, win_width=None, stdout_socket=None):

    def send_output(output):
        if stdout_socket:
            stdout_socket.send_multipart([
                b'PRINT',
                env.config.get('slave_id', '').encode(),
                output.encode()
            ])
        else:
            sys.stdout.write(output)

    if sys.platform == 'win32':
        import pexpect
        import pexpect.popen_spawn as ps
        child = ps.PopenSpawn(cmd)
        while True:
            try:
                child.expect('\n')
                if env.verbosity > 0:
                    send_output(child.before.decode() + '\n')
            except pexpect.EOF:
                break
        return child.wait()
    else:
        import pexpect
        import subprocess
        if win_width:
            os.environ['COLUMNS'] = str(win_width)
        else:
            os.environ['COLUMNS'] = '80'
        try:
            if isinstance(cmd, str):
                if shell:
                    child = pexpect.spawn(
                        '/bin/bash', ['-c', cmd], timeout=None)
                else:
                    child = pexpect.spawn(cmd, timeout=None)
            else:
                if shell:
                    child = pexpect.spawn(
                        '/bin/bash', ['-c', subprocess.list2cmdline(cmd)],
                        timeout=None)
                else:
                    child = pexpect.spawn(
                        subprocess.list2cmdline(cmd), timeout=None)
            while True:
                try:
                    child.expect('\r\n')
                    if env.verbosity > 0:
                        send_output(child.before.decode() + '\n')
                except pexpect.EOF:
                    break
            child.wait()
            child.close()
            return child.exitstatus
        except Exception as e:
            sys.stderr.write(str(e))
            return 1


def format_par(name, par):
    from .targets import path, file_target
    try:
        name = name.replace("_", "-")
        val = eval(par)
        # specify type
        if isinstance(val, type) or val is None:
            if val == bool:
                return f'--[no-]{name} (required)'
            elif val is None:
                return f'--{name} VAL (required)'
            elif val not in (str, file_target, path) and hasattr(
                    val, '__iter__'):
                return f'--{name} VAL VAL ... (as {val.__class__.__name__}, required)'
            else:
                return f'--{name} VAL (as {val.__name__}, required)'
        else:
            if val is True or val is False:
                return f'--[no-]{name} (default to {val})'
            elif isinstance(val, Sequence) and not isinstance(val, str):
                return f'--{name} {" ".join(str(x) for x in val)} (as {val.__class__.__name__})'
            elif isinstance(val, str):
                return f'--{name} {val if val.isalnum() else repr(val)}'
            else:
                return f'--{name} {val} (as {val.__class__.__name__})'
    except Exception:
        return f'--{name} {par}'


def b64_of(filename: str):
    with open(filename, 'rb') as content:
        data = content.read()
    return base64.b64encode(data).decode('ascii')


def dot_to_gif(filename: str, warn=None):
    import glob
    import tempfile
    from graphviz import Source
    from PIL import Image, ImageDraw, ImageFont
    with open(filename) as dot, tempfile.TemporaryDirectory() as tempDirectory:
        content = dot.read()
        # find out subworkflows
        subworkflows = [
            line.split()[2].strip('"')
            for line in content.splitlines()
            if line.startswith('strict digraph')
        ]
        # find out unique subgraphs
        unique_subworkflows = list(dict.fromkeys(subworkflows))
        #
        src = Source(content)
        src.format = 'png'
        outfile = src.render(filename='sosDot', directory=tempDirectory)
        # dot command can generate more than outfiles returned by the render function
        pngFiles = glob.glob(os.path.join(tempDirectory, f'sosDot*.png'))
        if len(pngFiles) == 1:
            return b64_of(outfile)
        # create a gif files from multiple png files
        pngFiles.sort(
            key=lambda x: int(os.path.basename(x)[:-3].split('.')[1] or 0))
        # find maximum size for all graphs corresponding to their subgraphs
        maxWidth = 0
        wf_maxHeight = {}
        images = {}
        for subworkflow in unique_subworkflows:
            wf_images = {
                png: Image.open(png)
                for png, wf in zip(pngFiles, subworkflows)
                if wf == subworkflow
            }
            maxWidth = max(maxWidth,
                           max([x.size[0] for x in wf_images.values()]))
            wf_maxHeight[subworkflow] = max(
                [x.size[1] for x in wf_images.values()]) + 20
            images.update(wf_images)
        # now, we stack workflows as follows
        #    G1
        #    G2
        #    G3
        # and allow G1, G2, G3 to expand...
        maxWidth += 150
        totalHeight = sum(wf_maxHeight.values())
        lastGraph = {}
        newImages = {}
        try:
            font = ImageFont.truetype('/Library/Fonts/Arial.ttf', 8)
        except Exception:
            try:
                font = ImageFont.truetype('arial.ttf', 8)
            except Exception:
                font = None

        for subworkflow, pngFile in zip(subworkflows, pngFiles):
            image = images[pngFile]
            lastGraph[subworkflow] = image
            # we need to stitch figures together
            try:
                newImg = Image.new(
                    "RGB", (maxWidth, totalHeight), color=0xFFFFFF)
                y_loc = 0
                for wf in unique_subworkflows:
                    # if current, use the new one
                    if wf == subworkflow:
                        img = image
                    elif wf in lastGraph:
                        img = lastGraph[wf]
                    else:
                        continue
                    newImg.paste(img,
                                 ((maxWidth - img.size[0]) // 2, y_loc + 20))
                    draw = ImageDraw.Draw(newImg)
                    # font = ImageFont.truetype("sans-serif.ttf", 8)
                    # , font=font)
                    draw.text((5, y_loc + 5), wf, (0, 0, 0), font=font)
                    y_loc += wf_maxHeight[wf]
            except Exception as e:
                if warn:
                    warn(f'Failed to resize gif file: {e}')
                return b64_of(pngFiles[-1])
            newImages[pngFile] = newImg

        # create a gif file from images
        try:
            with BytesIO() as output:
                newImages[pngFiles[0]].save(
                    output,
                    save_all=True,
                    format='GIF',
                    append_images=[newImages[x] for x in pngFiles[1:]],
                    duration=400,
                    loop=0)
                return base64.b64encode(output.getvalue()).decode('ascii')
        except Exception as e:
            if warn:
                warn(f'Failed to generate gif animation: {e}')
        # if things go wrong
        return b64_of(pngFiles[-1])


def separate_options(options: str) -> List[str]:
    pieces = options.split(',')
    idx = 0
    while True:
        try:
            # test current group
            compile(
                pieces[idx].strip(),
                filename='<string>',
                mode='exec' if '=' in pieces[idx] else 'eval')
            # if it is ok, go next
            idx += 1
            if idx == len(pieces):
                break
        except Exception:
            # error happens merge the next piece
            if idx < len(pieces) - 1:
                pieces[idx] += ',' + pieces[idx + 1]
                # error happens merge the next piece
                pieces.pop(idx + 1)
            else:
                # if no next group, expand previously correct one
                if idx == 0:
                    raise ValueError('Invalid section option')
                # break myself again
                pieces = pieces[: idx] + \
                    pieces[idx].split(',') + pieces[idx + 1:]
                # go back
                idx -= 1
                pieces[idx] += '\n' + pieces[idx + 1]
                pieces.pop(idx + 1)
    return pieces


def get_localhost_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP


#
# Handling -j option on cluster
#


def get_nodelist():
    """Return a list of hosts depending on the environment"""
    # If on a cluster, return some SoS acceptable format of node:nproc.
    if "SLURM_NODELIST" in os.environ:
        # Use scontrol utility to get the hosts list
        import subprocess
        hostsstr = subprocess.check_output(
            ["scontrol", "show", "hostnames",
             os.environ['SLURM_NODELIST']]).decode()
        # Split using endline
        # Split using endlinea, numer of cpus can be in the format of 4(x2)
        nprocs = os.environ["SLURM_JOB_CPUS_PER_NODE"].split('(')[0]
        args = [
            f'{x}:{nprocs}' for x in hostsstr.split(os.linesep) if x.strip()
        ]
        env.log_to_file('WORKER', f'Using "-j {args}" on a SLURM cluster.')
        return args
    elif "PBS_ENVIRONMENT" in os.environ:
        with open(os.environ["PBS_NODEFILE"], 'r') as hosts:
            hostlist = hosts.read().split()
            from collections import Counter, OrderedDict
            counts = Counter(hostlist)
            args = [
                f'{x}:{counts[x]}'
                for x in list(OrderedDict.fromkeys(hostlist))
            ]
        env.log_to_file('WORKER', f'Using "-j {args}" on a PBS cluster.')
        return args
    elif "PE_HOSTFILE" in os.environ:
        # The pathname of a file containing a  detailed  descrip-
        #  tion  of  the  layout of the parallel environment to be
        #  setup by the start-up procedure. Each line of the  file
        #  refers  to a host on which parallel processes are to be
        #  run. The first entry of each line denotes the hostname,
        #  the second entry the number of parallel processes to be
        #  run on the host, the third entry the name of the queue,
        #  and  the  fourth  entry a processor range to be used in
        #  case of a multiprocessor machine.
        with open(os.environ["PE_HOSTFILE"], 'r') as hosts:
            args = [':'.join(host.split()[:2]) for host in hosts]
        env.log_to_file('WORKER', f'Using "-j {args}" on a SGE cluster.')
        return args
    elif "LSB_MCPU_HOSTS" in os.environ:
        # IBM LSF https://www.ibm.com/support/knowledgecenter/en/SSETD4_9.1.3/lsf_config_ref/lsf_envars_ref.html
        # LSB_MCPU_HOSTS="hostA 3 hostB 3"
        hostlist = os.environ["LSB_MCPU_HOSTS"].strip().split()
        args = [
            f'{hostlist[2*i]}:{hostlist[2*i+1]}'
            for i in range(len(hostlist) // 2)
        ]
        env.log_to_file('WORKER', f'Using "-j {args}" on a IBM LSF cluster.')
        return args


def under_cluster():
    return any(
        x in os.environ for x in ("SLURM_NODELIST", "PBS_ENVIRONMENT",
                                  "PE_HOSTFILE", "LSB_MCPU_HOSTS"))


def get_open_files_and_connections(pid):
    proc = psutil.Process(pid)
    return {'open_files': proc.open_files(), 'connections': proc.connections()}



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import ast
import copy
import os
import subprocess
import sys
import time
from collections import defaultdict
from collections.abc import Mapping, Sequence
from typing import List

import zmq

from .controller import close_socket, create_socket, send_message_to_controller
from .messages import encode_msg, decode_msg
from .eval import SoS_eval, SoS_exec, accessed_vars, KeepOnlyImportAndDefine
from .executor_utils import (__named_output__, __null_func__, __output_from__,
                             __traced__, clear_output, create_task,
                             get_traceback_msg, reevaluate_output, statementMD5,
                             validate_step_sig, verify_input, ExecuteError)
from .syntax import (SOS_DEPENDS_OPTIONS, SOS_INPUT_OPTIONS, SOS_OUTPUT_OPTIONS,
                     SOS_TARGETS_OPTIONS)
from .targets import (RemovedTarget, RuntimeInfo, UnavailableLock, sos_variable,
                      UnknownTarget, dynamic, file_target, sos_step,
                      sos_targets, invalid_target, textMD5)
from .tasks import MasterTaskParams, TaskFile
from .utils import (ArgumentError, StopInputGroup, TerminateExecution, env,
                    get_traceback, short_repr, ProcessKilled, get_localhost_ip)

__all__: List = []


class TaskManager:
    # manage tasks created by the step
    def __init__(self, num_tasks, trunk_size, trunk_workers):
        super(TaskManager, self).__init__()
        self.num_tasks = num_tasks
        import math
        self._slots = [[] for x in range(math.ceil(num_tasks / trunk_size))]
        self._last_slot_size = trunk_size if (num_tasks %
                                              trunk_size == 0) else (num_tasks %
                                                                     trunk_size)
        self.trunk_size = trunk_size
        self.trunk_workers = trunk_workers

        self._submitted_tasks = []
        # entire groups
        self._unsubmitted_slots = []
        # collection of partial groups if some tasks are completed
        self._unsubmitted_tasks = []
        # derived from _unsubmitted_slots
        self._all_ids = []
        self._all_output = []
        #
        self._terminate = False
        #
        self._tags = {}

    def set(self, idx, task_def):
        slot = idx // self.trunk_size
        #
        # slot [
        #   [idx, None] <- for empty
        #   [idx, taskdef] <- for non empty
        #  ]
        self._slots[slot].append([idx, task_def])
        # the slot is full
        if len(self._slots[slot]) == self.trunk_size or \
            (slot == len(self._slots) - 1 and len(self._slots[slot]) == self._last_slot_size):
            # if there are valida tasks
            if not all([x[1] is None for x in self._slots[slot]]):
                # remove empty tasks and sort by id
                if self.trunk_size == 1 or any(
                        x[1] is None for x in self._slots[slot]):
                    # if partial, sent to partial list
                    self._unsubmitted_tasks.extend(
                        [x[1] for x in self._slots[slot] if x[1] is not None])
                else:
                    self._unsubmitted_slots.append(
                        sorted(self._slots[slot], key=lambda x: x[0]))
            # clear skit
            self._slots[slot] = []
        if not task_def:
            return
        if isinstance(task_def[2], Sequence):
            self._all_output.extend(task_def[2])
        self._all_ids.append(task_def[0])
        self._tags[task_def[0]] = task_def[1].tags

    def tags(self, task_id):
        return self._tags.get(task_id, [])

    def index_of(self, task_id):
        if task_id in self._all_ids:
            return self._all_ids.index(task_id)
        else:
            return -1

    def has_output(self, output):
        if not isinstance(output, Sequence) or not self._unsubmitted_slots:
            return False
        return any(x in self._all_output for x in output)

    def get_job(self, all_tasks=False):
        # single tasks
        ids = []
        # submit all tasks without trunk, easy
        for slot in self._unsubmitted_slots:
            # create a master task
            master = MasterTaskParams(self.trunk_workers)
            for _, (task_id, taskdef, _) in slot:
                master.push(task_id, taskdef)
            ids.append(master.ID)
            TaskFile(master.ID).save(master.finalize())
            send_message_to_controller([
                'workflow_sig', 'task', master.ID,
                f"{{'creation_time': {time.time()}}}"
            ])
        self._unsubmitted_slots = []

        # individual tasks...
        if self.trunk_size == 1 or all_tasks:
            to_be_submitted = self._unsubmitted_tasks
            [
                to_be_submitted.extend([x[1]
                                        for x in slot
                                        if x[1] is not None])
                for slot in self._slots
                if slot
            ]
            self._unsubmitted_tasks = []
        else:
            # save complete blocks
            num_tasks = len(
                self._unsubmitted_tasks) // self.trunk_size * self.trunk_size
            to_be_submitted = self._unsubmitted_tasks[:num_tasks]
            self._unsubmitted_tasks = self._unsubmitted_tasks[num_tasks:]

        if self.trunk_size == 1 or (all_tasks and
                                    len(self._unsubmitted_tasks) == 1):
            for task_id, taskdef, _ in to_be_submitted:
                # if the task file, perhaps it is already running, we do not change
                # the task file. Otherwise we are changing the status of the task
                TaskFile(task_id).save(taskdef)
                send_message_to_controller([
                    'workflow_sig', 'task', task_id,
                    f"{{'creation_time': {time.time()}}}"
                ])
                ids.append(task_id)
        else:
            master = None
            for task_id, taskdef, _ in to_be_submitted:
                if master is not None and master.num_tasks() == self.trunk_size:
                    ids.append(master.ID)
                    TaskFile(master.ID).save(master)
                    send_message_to_controller([
                        'workflow_sig', 'task', master.ID,
                        f"{{'creation_time': {time.time()}}}"
                    ])
                    master = None
                if master is None:
                    master = MasterTaskParams(self.trunk_workers)
                master.push(task_id, taskdef)
            # the last piece
            if master is not None:
                TaskFile(master.ID).save(master.finalize())
                send_message_to_controller([
                    'workflow_sig', 'task', master.ID,
                    f"{{'creation_time': {time.time()}}}"
                ])
                ids.append(master.ID)

        if not ids:
            return None

        self._submitted_tasks.extend(ids)
        return ids

    def clear_submitted(self):
        self._submitted_tasks = []


def expand_input_files(*args, **kwargs):
    # if unspecified, use __step_output__ as input (default)
    # resolve dynamic input.
    args = [x.resolve() if isinstance(x, dynamic) else x for x in args]
    kwargs = {
        x: (y.resolve() if isinstance(y, dynamic) else y)
        for x, y in kwargs.items()
    }

    # if no input,
    if not args and not kwargs:
        return env.sos_dict['step_input']
    # if only group_by ...
    elif not args and all(x in SOS_TARGETS_OPTIONS for x in kwargs.keys()):
        return sos_targets(
            env.sos_dict['step_input'],
            _verify_existence=env.config['error_mode'] != 'ignore',
            **kwargs)
    else:
        return sos_targets(
            *args,
            **kwargs,
            _verify_existence=env.config['error_mode'] != 'ignore',
            _undetermined=False,
            _source=env.sos_dict['step_name'])


def expand_depends_files(*args, **kwargs):
    '''handle directive depends'''
    args = [x.resolve() if isinstance(x, dynamic) else x for x in args]
    kwargs = {
        x: (y.resolve() if isinstance(y, dynamic) else y)
        for x, y in kwargs.items()
    }
    return sos_targets(
        *args,
        **kwargs,
        _verify_existence=True,
        _undetermined=False,
        _source=env.sos_dict['step_name'])


def expand_output_files(value, *args, **kwargs):
    '''Process output files (perhaps a pattern) to determine input files.
    '''
    if any(isinstance(x, dynamic) for x in args) or any(
            isinstance(y, dynamic) for y in kwargs.values()):
        return sos_targets(_undetermined=value)
    else:
        return sos_targets(
            *args,
            **kwargs,
            _undetermined=False,
            _source=env.sos_dict['step_name'])


def parse_shared_vars(option):
    shared_vars = set()
    if not option:
        return shared_vars
    if isinstance(option, str):
        shared_vars.add(option)
    elif isinstance(option, Mapping):
        for val in option.values():
            shared_vars |= accessed_vars(val, mode='eval')
    elif isinstance(option, Sequence):
        for item in option:
            if isinstance(item, str):
                shared_vars.add(item)
            elif isinstance(item, Mapping):
                for val in item.values():
                    shared_vars |= accessed_vars(val, mode='eval')
    return shared_vars


def evaluate_shared(vars, option):
    # handle option shared and store variables in a "__shared_vars" variable
    shared_vars = {}
    env.sos_dict.quick_update(vars[-1])
    for key in vars[-1].keys():
        try:
            if key in ('output', 'depends', 'input'):
                env.logger.warning(
                    f'Cannot overwrite variable step_{key} from substep variable {key}'
                )
            else:
                env.sos_dict.set('step_' + key, [x[key] for x in vars])
        except Exception as e:
            env.logger.warning(
                f'Failed to create step level variable step_{key}: {e}')
    if isinstance(option, str):
        if option in env.sos_dict:
            shared_vars[option] = env.sos_dict[option]
        else:
            raise RuntimeError(f'shared variable does not exist: {option}')
    elif isinstance(option, Mapping):
        for var, val in option.items():
            try:
                if var == val:
                    shared_vars[var] = env.sos_dict[var]
                else:
                    shared_vars[var] = SoS_eval(val)
            except Exception as e:
                raise RuntimeError(
                    f'Failed to evaluate shared variable {var} from expression {val}: {e}'
                )
    # if there are dictionaries in the sequence, e.g.
    # shared=['A', 'B', {'C':'D"}]
    elif isinstance(option, Sequence):
        for item in option:
            if isinstance(item, str):
                if item in env.sos_dict:
                    shared_vars[item] = env.sos_dict[item]
                else:
                    raise RuntimeError(
                        f'shared variable does not exist: {option}')
            elif isinstance(item, Mapping):
                for var, val in item.items():
                    try:
                        if var == val:
                            continue
                        else:
                            shared_vars[var] = SoS_eval(val)
                    except Exception as e:
                        raise RuntimeError(
                            f'Failed to evaluate shared variable {var} from expression {val}: {e}'
                        )
            else:
                raise RuntimeError(
                    f'Unacceptable shared option. Only str or mapping are accepted in sequence: {option}'
                )
    else:
        raise RuntimeError(
            f'Unacceptable shared option. Only str, sequence, or mapping are accepted in sequence: {option}'
        )
    return shared_vars


def get_value_of_param(name, param_list, extra_dict={}):
    tree = ast.parse(f'__null_func__({param_list})')
    # x.func can be an attribute (e.g. a.b()) and do not have id
    kwargs = [
        x for x in ast.walk(tree)
        if x.__class__.__name__ == 'keyword' and x.arg == name
    ]
    if not kwargs:
        return []
    try:
        return [ast.literal_eval(kwargs[0].value)]
    except Exception:
        return [
            eval(
                compile(
                    ast.Expression(body=kwargs[0].value),
                    filename='<string>',
                    mode="eval"), extra_dict)
        ]


def is_sos_run_the_only_last_stmt(stmt):
    tree = ast.parse(stmt)
    return len(tree.body) >= 1 and \
        isinstance(tree.body[-1], ast.Expr) and \
        isinstance(tree.body[-1].value, ast.Call) and \
        hasattr(tree.body[-1].value.func, 'id') and \
        tree.body[-1].value.func.id == 'sos_run' and \
        len([x for x in ast.walk(tree) if isinstance(x, ast.Call) and hasattr(x.func, 'id') and x.func.id == 'sos_run']) == 1


class Base_Step_Executor:
    # This base class defines how steps are executed. The derived classes will reimplement
    # some function to behave differently in different modes.
    #
    def __init__(self, step):
        self.step = step
        self.task_manager = None
        self.exec_error = ExecuteError(self.step.step_name())

    #
    #  Functions that should be redefined in derived class
    #

    def submit_tasks(self, tasks):
        raise RuntimeError('Undefined base function submit_tasks')

    def wait_for_tasks(self, tasks, all_submitted):
        # this will be redefined in subclasses
        raise RuntimeError('Undefined base function wait_for_tasks')

    def wait_for_subworkflows(self, allow_pending=0):
        raise RuntimeError('Undefined base function wait_for_subworkflows')

    def handle_unknown_target(self, e):
        raise RuntimeError('Undefined base function handle_unknown_target')

    def init_input_output_vars(self):
        # if there is __step_output__ from previous step, use it as default input
        # otherwise, reset to empty
        if '__step_output__' not in env.sos_dict or env.sos_dict[
                '__step_output__'].unspecified():
            env.sos_dict.set('step_input', sos_targets([]))
        else:
            env.sos_dict.set('step_input', env.sos_dict['__step_output__'])
        # input can be Undetermined from undetermined output from last step
        env.sos_dict.set('_input', copy.deepcopy(env.sos_dict['step_input']))

        # if there is default output for auxiliary steps, use it as step_output and _output
        # otherwise reset to unspecified.
        if '__default_output__' in env.sos_dict:
            # if step is triggered by sos_step, it should not be considered as
            # output of the step. #981
            env.sos_dict.set(
                '__default_output__',
                sos_targets([
                    x for x in env.sos_dict['__default_output__']._targets
                    if not isinstance(x, sos_step)
                ]))
            env.sos_dict.set('step_output',
                             copy.deepcopy(env.sos_dict['__default_output__']))
            env.sos_dict.set('_output',
                             copy.deepcopy(env.sos_dict['__default_output__']))
        else:
            env.sos_dict.set('step_output', sos_targets([]))
            # output is said to be unspecified until output: is used
            env.sos_dict.set('_output', sos_targets(_undetermined=True))

        env.sos_dict.set('step_depends', sos_targets([]))
        env.sos_dict.set('_depends', sos_targets([]))

    #
    # Common functions
    #

    def verify_output(self):
        missing = sos_targets([])
        if env.sos_dict['step_output'] is None:
            return
        if not env.sos_dict['step_output'].valid():
            raise RuntimeError(
                'Output of a completed step cannot be undetermined or unspecified.'
            )
        for target in env.sos_dict['step_output']:
            if isinstance(target, (sos_step, invalid_target)):
                continue
            if isinstance(target, str):
                if not file_target(target).target_exists('any'):
                    if env.config['run_mode'] == 'dryrun':
                        # in dryrun mode, we just create these targets
                        file_target(target).create_placeholder()
                    else:
                        # latency wait for 2 seconds because the file system might be slow
                        if env.config['run_mode'] == 'run':
                            time.sleep(2)
                        if not file_target(target).target_exists('any'):
                            if env.config['error_mode'] == 'ignore':
                                missing.extend(target)
                            else:
                                raise RuntimeError(
                                    f'Output target {target} does not exist after the completion of step {env.sos_dict["step_name"]} (curdir={os.getcwd()})'
                                )
            elif not target.target_exists('any'):
                if env.config['run_mode'] == 'dryrun':
                    target.create_placeholder()
                else:
                    if env.config['run_mode'] == 'run':
                        time.sleep(2)
                    if not target.target_exists('any'):
                        if env.config['error_mode'] == 'ignore':
                            missing.extend(target)
                        else:
                            raise RuntimeError(
                                f'Output target {target} does not exist after the completion of step {env.sos_dict["step_name"]}'
                            )
        return missing

    # directive input
    def process_input_args(self, ifiles: sos_targets, **kwargs):
        """This function handles directive input and all its parameters.
        It
            determines and set __step_input__
            determines and set pattern variables if needed
        returns
            _groups
            _vars
        which are groups of _input and related _vars
        """
        if ifiles.unspecified():
            env.sos_dict.set('step_input', sos_targets([]))
            env.sos_dict.set('_input', sos_targets([]))
            env.sos_dict.set('step_output', sos_targets())
            return [sos_targets([])], [{}]

        assert isinstance(ifiles, sos_targets)

        if env.sos_dict.get('__dynamic_input__', False):
            runner = self.verify_dynamic_targets(
                [x for x in ifiles if isinstance(x, file_target)])
            try:
                yreq = next(runner)
                while True:
                    yres = yield yreq
                    yreq = runner.send(yres)
            except StopIteration:
                pass

        # input file is the filtered files
        env.sos_dict.set('step_input', ifiles)
        env.sos_dict.set('_input', ifiles)

        if ifiles._num_groups() == 0:
            ifiles._group('all')
        #
        return ifiles.groups

    def verify_dynamic_targets(self, target):
        yield None
        return True

    def process_depends_args(self, dfiles: sos_targets, **kwargs):
        for k in kwargs.keys():
            if k not in SOS_DEPENDS_OPTIONS:
                raise RuntimeError(f'Unrecognized depends option {k}')
        if dfiles.undetermined():
            raise ValueError(r"Depends needs to handle undetermined")

        if env.sos_dict.get('__dynamic_depends__', False):
            runner = self.verify_dynamic_targets(
                [x for x in dfiles if isinstance(x, file_target)])
            try:
                yreq = next(runner)
                while True:
                    yres = yield yreq
                    yreq = runner.send(yres)
            except StopIteration:
                pass

        env.sos_dict.set('_depends', dfiles)
        env.sos_dict.set('step_depends', dfiles)

    def process_output_args(self, ofiles: sos_targets, **kwargs):
        for k in kwargs.keys():
            if k not in SOS_OUTPUT_OPTIONS:
                raise RuntimeError(f'Unrecognized output option {k}')

        if ofiles._num_groups() > 0:
            if ofiles._num_groups() == 1:
                ofiles = ofiles._get_group(0)
            elif ofiles._num_groups() != len(self._substeps):
                raise RuntimeError(
                    f'Inconsistent number of output ({ofiles._num_groups()}) and input ({len(self._substeps)}) groups.'
                )
            else:
                ofiles = ofiles._get_group(env.sos_dict['_index'])

        # create directory
        if ofiles.valid():
            parents = set([
                os.path.abspath(os.path.join(ofile, os.pardir))
                for ofile in ofiles
                if isinstance(ofile, file_target)
            ])
            for parent_dir in parents:
                if parent_dir and not os.path.isdir(parent_dir):
                    os.makedirs(parent_dir, exist_ok=True)

        # set variables
        env.sos_dict.set('_output', ofiles)
        env.sos_dict.set('step_output', ofiles)
        #
        for ofile in ofiles:
            oname = ofile.target_name()
            if oname in self._all_outputs:
                raise ValueError(
                    f'Output {ofile} from substep {env.sos_dict["_index"]} of {env.sos_dict["__num_groups__"]} substeps overlaps with output from a previous substep.'
                )
            self._all_outputs.add(oname)

    def submit_task(self, task_info):
        if self.task_manager is None:
            if self.step.task_params:
                for key in ('trunk_size', 'trunk_workers', 'queue'):
                    val = get_value_of_param(
                        key,
                        self.step.task_params,
                        extra_dict=env.sos_dict.dict())
                    if val:
                        env.sos_dict['_runtime'][key] = val[0]

            if 'trunk_size' in env.sos_dict['_runtime']:
                trunk_size = env.sos_dict['_runtime']['trunk_size']
                if trunk_size is None or trunk_size <= 0:
                    trunk_size = env.sos_dict["__num_groups__"]
                if not isinstance(trunk_size, int):
                    raise ValueError(
                        f'An integer value or None is expected for runtime option trunk_size, "{trunk_size}" provided'
                    )
            else:
                trunk_size = 1
            if 'trunk_workers' in env.sos_dict['_runtime']:
                if 'nodes' in env.sos_dict['_runtime']:
                    raise ValueError(
                        'Option "trunk_workers" that specifies number of nodes and processes for the execution '
                        'of single-node jobs and option "nodes" that specifies number of nodes for single multi-node '
                        'jobs cannot be used at the same time.')
                trunk_workers = env.sos_dict['_runtime']['trunk_workers']
            else:
                trunk_workers = None

            # if 'queue' in env.sos_dict['_runtime'] and env.sos_dict['_runtime']['queue']:
            #    host = env.sos_dict['_runtime']['queue']
            # else:
            #    # otherwise, use workflow default
            #    host = '__default__'
            self.task_manager = TaskManager(env.sos_dict['__num_groups__'],
                                            trunk_size, trunk_workers)

        task_id = task_info['task_id']
        task_index = task_info['index']
        if task_id is None:
            self.task_manager.set(task_index, None)
            return None

        taskdef = task_info['task_def']
        task_vars = task_info['task_vars']

        # 618
        # it is possible that identical tasks are executed (with different underlying random numbers)
        # we should either give a warning or produce different ids...
        if self.task_manager.index_of(task_id) >= 0:
            raise RuntimeError(
                f'Task {task_id} generated for (_index={env.sos_dict["_index"]}) is identical to a previous one (_index={self.task_manager.index_of(task_id)}).'
            )
        elif self.task_manager.has_output(task_vars['_output']):
            raise RuntimeError(
                f'Task produces output files {", ".join(task_vars["_output"])} that are output of other tasks.'
            )
        # if no trunk_size, the job will be submitted immediately
        # otherwise tasks will be accumulated and submitted in batch
        self.task_manager.set(task_index,
                              (task_id, taskdef, task_vars['_output']))
        tasks = self.task_manager.get_job()
        if tasks:
            self.submit_tasks(tasks)
        return task_id

    def wait_for_results(self, all_submitted):
        # this is a generator function because wait_for_tasks is a generator
        # function and needs to yield to the caller
        if self.concurrent_substep:
            try:
                runner = self.wait_for_substep()
                yreq = next(runner)
                while True:
                    yres = yield yreq
                    yreq = runner.send(yres)
            except StopIteration:
                pass

        if self.task_manager is None:
            return {}

        #
        # report task
        # what we should do here is to get the alias of the Host
        # because it can be different (e.g. not localhost
        queue = env.sos_dict['_runtime']['queue']

        # submit the last batch of tasks
        tasks = self.task_manager.get_job(all_tasks=True)
        if tasks:
            self.submit_tasks(tasks)

        # waiting for results of specified IDs
        try:
            #1218
            runner = self.wait_for_tasks(self.task_manager._submitted_tasks,
                                         all_submitted)
            yreq = next(runner)
            while True:
                yres = yield yreq
                yreq = runner.send(yres)
        except StopIteration as e:
            results = e.value

        for id, result in results.items():
            # turn to string to avoid naming lookup issue
            rep_result = {
                x: (y if isinstance(y,
                                    (int, bool, float, str)) else short_repr(y))
                for x, y in result.items()
            }
            rep_result['tags'] = ' '.join(self.task_manager.tags(id))
            rep_result['queue'] = queue
            send_message_to_controller(
                ['workflow_sig', 'task', id,
                 repr(rep_result)])
        self.task_manager.clear_submitted()

        # if in dryrun mode, we display the output of the dryrun task
        if env.config['run_mode'] == 'dryrun':
            tid = list(results.keys())[0]
            tf = TaskFile(tid)
            if tf.has_stdout():
                print(TaskFile(tid).stdout)

        for idx, task in self.proc_results.items():
            # if it is done
            if isinstance(task, dict):
                continue
            if task in results:
                self.proc_results[idx] = results[task]
            else:
                # can be a subtask
                for _, mres in results.items():
                    if 'subtasks' in mres and task in mres['subtasks']:
                        self.proc_results[idx] = mres['subtasks'][task]
                    #elif 'exception' in mres:
                    #    self.proc_results[idx] = mres
        #
        # check if all have results?
        if any(isinstance(x, str) for x in self.proc_results.values()):
            raise RuntimeError(
                f'Failed to get results for tasks {", ".join(x for x in self.proc_results.values() if isinstance(x, str))}'
            )
        #
        for idx, res in self.proc_results.items():
            if 'skipped' in res and res['skipped']:
                self.completed['__task_skipped__'] += 1
                # complete case: task skipped
                send_message_to_controller(
                    ['progress', 'substep_completed', env.sos_dict['step_id']])
            else:
                # complete case: task completed
                send_message_to_controller(
                    ['progress', 'substep_ignored', env.sos_dict['step_id']])
                self.completed['__task_completed__'] += 1
            if 'shared' in res:
                self.shared_vars[idx].update(res['shared'])

    def log(self, stage=None, msg=''):
        if stage == 'start':
            env.logger.info(
                f'{"Checking" if env.config["run_mode"] == "dryrun" else "Running"} ``{self.step.step_name(True)}``: {self.step.comment.strip()}'
            )
        elif stage == 'input statement':
            if 'STEP' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                    'SOS_DEBUG']:
                env.log_to_file('STEP', f'Handling input statement {msg}')
        elif stage == '_input':
            if env.sos_dict['_input'] is not None and len(
                    env.sos_dict['_input']) > 0:
                env.logger.debug(
                    f'_input: ``{short_repr(env.sos_dict["_input"])}``{msg}')
        elif stage == '_depends':
            if env.sos_dict['_depends'] is not None:
                env.logger.debug(
                    f'_depends: ``{short_repr(env.sos_dict["_depends"])}``{msg}'
                )
        elif stage == 'input':
            if env.sos_dict['step_input'] is not None:
                env.logger.info(
                    f'input:   ``{short_repr(env.sos_dict["step_input"])}``{msg}'
                )
        elif stage == 'output':
            if env.sos_dict['step_output'] is not None and len(
                    env.sos_dict['step_output']) > 0:
                env.logger.info(
                    f'``{self.step.step_name(True)}`` output:   ``{short_repr(env.sos_dict["step_output"])}``{msg}'
                )

    def execute(self, stmt, return_result=False):
        try:
            self.last_res = SoS_exec(
                stmt,
                return_result=return_result or
                env.config['run_mode'] == 'interactive')
            if return_result:
                return self.last_res
        except (StopInputGroup, TerminateExecution, UnavailableLock):
            raise
        except subprocess.CalledProcessError as e:
            raise RuntimeError(e.stderr)
        except ArgumentError:
            raise
        except ProcessKilled:
            raise
        except KeyboardInterrupt as e:
            raise RuntimeError(get_traceback_msg(e))
        except Exception as e:
            raise RuntimeError(get_traceback_msg(e))

    def prepare_substep(self):
        # socket to collect result
        self.result_pull_socket = create_socket(env.zmq_context, zmq.PULL,
                                                'substep result collector')
        local_ip = get_localhost_ip()
        port = self.result_pull_socket.bind_to_random_port(f'tcp://{local_ip}')
        env.config['sockets']['result_push_socket'] = f'tcp://{local_ip}:{port}'

    def submit_substep(self, param):
        send_message_to_controller(['substep', param])

    def process_returned_substep_result(self, till=None, wait=True):
        while True:
            if not wait:
                # 1213
                cur_index = env.sos_dict['_index']
                pending_substeps = cur_index - self._completed_concurrent_substeps + 1
                if pending_substeps < (100 if isinstance(
                        self.concurrent_substep, bool) else
                                       self.concurrent_substep):
                    if not self.result_pull_socket.poll(0):
                        return
                elif 'STEP' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                        'SOS_DEBUG']:
                    # if there are more than 100 pending substeps
                    # we wait indefinitely for the results
                    env.log_to_file(
                        'STEP',
                        f'Wait for more substeps to be done before submitting. (index={cur_index}, processed={self._completed_concurrent_substeps})'
                    )
            elif self._completed_concurrent_substeps == till:
                return
            yield self.result_pull_socket
            res = decode_msg(self.result_pull_socket.recv())
            if 'exception' in res:
                if isinstance(res['exception'], ProcessKilled):
                    raise res['exception']
                elif isinstance(res['exception'], RemovedTarget):
                    pass
                elif env.config['error_mode'] == 'ignore':
                    idx_msg = f'(id={env.sos_dict["step_id"]}, index={res["index"]})' if "index" in res and len(
                        self._substeps
                    ) > 1 else f'(id={env.sos_dict["step_id"]})'
                    env.logger.warning(
                        f'''Ignoring error from ``{self.step.step_name(True)}`` {idx_msg}: {res["exception"]}.'''
                    )
                    res['output'] = sos_targets(invalid_target())
                elif env.config['error_mode'] == 'abort':
                    idx_msg = f'(id={env.sos_dict["step_id"]}, index={res["index"]})' if "index" in res and len(
                        self._substeps
                    ) > 1 else f'(id={env.sos_dict["step_id"]})'
                    self.exec_error.append(idx_msg, res['exception'])
                    # try to stop everything but wait till for submitted tasks to
                    # complete
                    self._completed_concurrent_substeps + 1
                    waiting = till - 1 - self._completed_concurrent_substeps
                    env.logger.warning(
                        f'``{self.step.step_name(True)}`` {idx_msg} returns an error.{f" Terminating step after completing {waiting} submitted substeps." if waiting else " Terminating now."}'
                    )
                    for i in range(waiting):
                        yield self.result_pull_socket
                        res = decode_msg(self.result_pull_socket.recv())
                        if 'exception' in res:
                            self.exec_error.append(f'index={res["index"]}',
                                                   res['exception'])
                    raise self.exec_error
                else:
                    # default or unspecified
                    idx_msg = f'(id={env.sos_dict["step_id"]}, index={res["index"]})' if "index" in res and len(
                        self._substeps
                    ) > 1 else f'(id={env.sos_dict["step_id"]})'
                    self.exec_error.append(idx_msg, res['exception'])
            #
            if "index" not in res:
                raise RuntimeError(
                    "Result received from substep does not have key index")
            if 'task_id' in res:
                task = self.submit_task(res)
                # if substep returns tasks, ...
                if res['task_id']:
                    self.proc_results[res['index']] = task
                else:
                    # if there is no task_id, the substep must have
                    # been skipped.
                    self.proc_results[res['index']] = res
            else:
                self.proc_results[res['index']] = res
            self._completed_concurrent_substeps += 1

    def wait_for_substep(self):
        while self._completed_concurrent_substeps < len(self.proc_results):
            try:
                runner = self.process_returned_substep_result(
                    till=len(self.proc_results), wait=True)
                yreq = next(runner)
                while True:
                    yres = yield yreq
                    yreq = runner.send(yres)
            except StopIteration:
                pass

    def collect_result(self):
        # only results will be sent back to the master process
        #
        # __step_input__:    input of this step
        # __steo_output__:   output of this step
        # __step_depends__:  dependent files of this step

        result = {
            '__step_input__': env.sos_dict['step_input'],
            '__step_output__': env.sos_dict['step_output'],
            '__step_depends__': env.sos_dict['step_depends'],
            '__step_name__': env.sos_dict['step_name'],
            '__completed__': self.completed,
        }
        result['__last_res__'] = self.last_res
        result['__shared__'] = {}
        if 'shared' in self.step.options:
            result['__shared__'] = self.shared_vars
        for x in result['__step_output__'].targets:
            if isinstance(x, sos_variable):
                result['__shared__'][x.target_name()] = env.sos_dict[
                    x.target_name()]
        send_message_to_controller([
            'progress', 'step_completed',
            -1 if 'sos_run' in env.sos_dict['__signature_vars__'] else
            self.completed['__step_completed__'], env.sos_dict['step_name'],
            env.sos_dict['step_output']
        ])
        return result

    def set_task_queue_from_task_params(self):
        if self.step.task_params:
            try:
                task_queue = get_value_of_param(
                    'queue',
                    self.step.task_params,
                    extra_dict=env.sos_dict.dict())
                if task_queue:
                    env.sos_dict['_runtime']['queue'] = task_queue[0]
            except Exception as e:
                raise ValueError(
                    f'Failed to determine value of parameter queue of {self.step.task_params}: {e}'
                )
            # # check concurrent #1134
            # try:
            #     task_concurrency = get_value_of_param(
            #         'concurrent',
            #         self.step.task_params,
            #         extra_dict=env.sos_dict.dict())
            #     if task_concurrency:
            #         env.sos_dict['_runtime']['concurrent'] = task_concurrency[0]
            # except Exception as e:
            #     raise ValueError(
            #         f'Failed to determine value of parameter queue of {self.step.task_params}: {e}'
            #     )
        # if -q is unspecified and option queue is unspecified,
        # or queue=None is specified, disregard the task keyword
        if (env.config['default_queue'] is None and
            'queue' not in env.sos_dict['_runtime']) or \
            ('queue' in env.sos_dict['_runtime'] and
            env.sos_dict['_runtime']['queue'] is None):
            # remove task statement
            if len(self.step.statements
                  ) >= 1 and self.step.statements[-1][0] == '!':
                self.step.statements[-1][1] += '\n' + self.step.task
            else:
                self.step.statements.append(['!', self.step.task])
            self.step.task = None
        # is queue is unspecified, it take value from command line
        # in this case -q should have been specified
        elif 'queue' not in env.sos_dict['_runtime']:
            env.sos_dict['_runtime']['queue'] = env.config['default_queue']

    def local_exec_without_signature(self, statement):
        idx = env.sos_dict["_index"]
        env.log_to_file(
            'STEP',
            f'Execute substep {env.sos_dict["step_name"]} without signature')
        try:
            if self.is_input_verified:
                verify_input()
                self.is_input_verified = False
            if env.sos_dict.get('__concurrent_subworkflow__', False):
                self._subworkflow_results.append(
                    self.execute(statement[1], return_result=True))
            else:
                self.execute(statement[1])
            if not self.step.task and env.config['run_mode'] != 'interactive':
                env.logger.info(
                    f'``{self.step.step_name(True)}``{f" (index={idx})" if len(self._substeps) > 1 else ""} is ``completed``{" (pending nested workflow)" if self._subworkflow_results else ""}.'
                )
        finally:
            if not self.step.task:
                # if no task, this step is __completed
                # complete case: local skip without task
                send_message_to_controller(
                    ['progress', 'substep_completed', env.sos_dict['step_id']])
        if 'shared' in self.step.options:
            try:
                self.shared_vars[env.sos_dict['_index']].update({
                    x: env.sos_dict[x]
                    for x in self.vars_to_be_shared
                    if x in env.sos_dict
                })
            except Exception as e:
                raise ValueError(f'Missing shared variable {e}.')

    def local_exec_with_signature(self, statement, sig):
        idx = env.sos_dict["_index"]
        # signature might be built outside of the function
        # not in a debug mode delayed to now
        if sig is None:
            sig = RuntimeInfo(
                statementMD5([statement[1], self.step.task]),
                env.sos_dict['_input'],
                env.sos_dict['_output'],
                env.sos_dict['_depends'],
                env.sos_dict['__signature_vars__'],
                shared_vars=self.vars_to_be_shared)
            # if singaure match, we skip the substep even  if
            # there are tasks.
            matched = validate_step_sig(sig)
            if matched:
                if env.sos_dict['step_output'].undetermined():
                    self.output_groups[idx] = matched["output"]
                if 'vars' in matched:
                    self.shared_vars[idx].update(matched["vars"])
                return True

        env.log_to_file(
            'STEP',
            f'Execute substep {env.sos_dict["step_name"]} with signature {sig.sig_id}'
        )
        sig.lock()
        try:
            if self.is_input_verified:
                verify_input()
                self.is_input_verified = False
            if env.sos_dict.get('__concurrent_subworkflow__', False):
                self._subworkflow_results.append(
                    self.execute(statement[1], return_result=True))
            else:
                self.execute(statement[1])
            if not self.step.task and env.config['run_mode'] != 'interactive':
                env.logger.info(
                    f'``{self.step.step_name(True)}``{f" (index={idx})" if len(self._substeps) > 1 else ""} is ``completed``{" (pending nested workflow)" if self._subworkflow_results else ""}.'
                )
            if 'shared' in self.step.options:
                try:
                    self.shared_vars[env.sos_dict['_index']].update({
                        x: env.sos_dict[x]
                        for x in self.vars_to_be_shared
                        if x in env.sos_dict
                    })
                except Exception as e:
                    raise ValueError(f'Missing shared variable {e}.')
        finally:
            # if this is the end of substep, save the signature
            # otherwise we need to wait for the completion
            # of the task.
            if not self.step.task:
                if env.sos_dict['step_output'].undetermined():
                    output = reevaluate_output()
                    self.output_groups[env.sos_dict['_index']] = output
                    sig.set_output(output)
                sig.write()
                # complete case : local execution without task
                send_message_to_controller(
                    ['progress', 'substep_completed', env.sos_dict['step_id']])
            else:
                self.pending_signatures[idx] = sig
            sig.release()
        return False

    def skip_substep(self):
        idx = env.sos_dict["_index"]
        # if concurrent substep, there might be later steps that needs to be rerun
        # and we need to mark some steps has been completed.
        if self.concurrent_substep:
            self._completed_concurrent_substeps += 1
            self.proc_results[idx] = {
                'index': idx,
                'ret_code': 0,
                'output': copy.deepcopy(env.sos_dict['_output'])
            }
        send_message_to_controller(
            ['progress', 'substep_ignored', env.sos_dict['step_id']])

    def concurrent_exec(self, statement, sig=None):
        idx = env.sos_dict["_index"]
        env.log_to_file(
            'STEP',
            f'Execute substep {env.sos_dict["step_name"]} {idx} concurrently with {self._completed_concurrent_substeps} completed'
        )

        # the ignatures are supposed to be written by substep worker, however
        # the substep worker might send tasks back to the step worker and
        # we should write the signatures after the tasks are completed
        if env.config['sig_mode'] != 'ignore' and not env.sos_dict['_output'].unspecified() \
            and self.step.task:
            self.pending_signatures[idx] = sig if sig else RuntimeInfo(
                statementMD5([statement[1], self.step.task]),
                env.sos_dict['_input'],
                env.sos_dict['_output'],
                env.sos_dict['_depends'],
                env.sos_dict['__signature_vars__'],
                shared_vars=self.vars_to_be_shared)
        #
        # step_output: needed only when it is undetermined
        # step_input: not needed
        # _input, _output, _depends, _index: needed
        # step_name: for debug scripts
        # step_id, workflow_id: for reporting to controller
        # '__signature_vars__' to be used for signature creation
        #
        # __step_context__ is not needed because substep
        # executor does not support nested workflow

        proc_vars = env.sos_dict['__signature_vars__'] \
            | env.sos_dict['__environ_vars__'] | {'_input', '_output', '_depends', '_index',
             'step_output', 'step_name',
              '_runtime', 'step_id', 'workflow_id', '__num_groups__',
              '__signature_vars__'}
        self.proc_results[env.sos_dict['_index']] = {}
        self.submit_substep(
            dict(
                stmt=statement[1],
                global_def=self.step.global_def,
                #1225: the step might contain large variables from global section, but
                # we do not have to sent them if they are not used in substeps.
                global_vars={
                    x: y
                    for x, y in self.step.global_vars.items()
                    if x in env.sos_dict['__signature_vars__']
                },
                task=self.step.task,
                task_params=self.step.task_params,
                proc_vars=env.sos_dict.clone_selected_vars(proc_vars),
                shared_vars=self.vars_to_be_shared,
                config=env.config))

    def check_task_sig(self):
        idx = env.sos_dict["_index"]
        sig = RuntimeInfo(
            statementMD5([self.step.task]),
            env.sos_dict['_input'],
            env.sos_dict['_output'],
            env.sos_dict['_depends'],
            env.sos_dict['__signature_vars__'],
            shared_vars=self.vars_to_be_shared)
        env.log_to_file(
            'STEP',
            f'Check task-only step {env.sos_dict["step_name"]} with signature {sig.sig_id}'
        )
        matched = validate_step_sig(sig)
        skip_index = bool(matched)
        if matched:
            if env.sos_dict['step_output'].undetermined():
                self.output_groups[env.sos_dict['_index']] = matched["output"]
            self.shared_vars[env.sos_dict['_index']].update(matched["vars"])
            # complete case: step with task ignored
            send_message_to_controller(
                ['progress', 'substep_ignored', env.sos_dict['step_id']])
        self.pending_signatures[idx] = sig
        return skip_index

    # def is_task_active(self):
    #     active = env.sos_dict['_runtime']['active']
    #     env.logger.error(active)
    #     if active is True:
    #         return True
    #     elif active is False:
    #         return False
    #     elif isinstance(active, int):
    #         if active >= 0 and env.sos_dict['_index'] != active:
    #             return False
    #         if active < 0 and env.sos_dict[
    #                 '_index'] != active + env.sos_dict['__num_groups__']:
    #             return False
    #         return True
    #     elif isinstance(active, Sequence):
    #         allowed_index = list([
    #             x if x >= 0 else env.sos_dict['__num_groups__'] + x
    #             for x in active
    #         ])
    #         return env.sos_dict['_index'] in allowed_index
    #     elif isinstance(active, slice):
    #         allowed_index = list(range(env.sos_dict['__num_groups__']))[active]
    #         return env.sos_dict['_index'] in allowed_index
    #     else:
    #         raise RuntimeError(
    #             f'Unacceptable value for option active: {active}')

    def check_results(self):
        for proc_result in [x for x in self.proc_results.values() if x['ret_code'] == 0]:
            if 'stdout' in proc_result and proc_result['stdout']:
                sys.stdout.write(proc_result['stdout'])
            if 'stderr' in proc_result and proc_result['stderr']:
                sys.stderr.write(proc_result['stderr'])

        # now that output is settled, we can write remaining signatures
        for idx, res in self.proc_results.items():
            if self.pending_signatures[idx] is not None and res[
                    'ret_code'] == 0 and 'sig_skipped' not in res:
                # task might return output with vars #1355
                self.pending_signatures[idx].set_output(self.output_groups[idx])
                self.pending_signatures[idx].write()
            if res['ret_code'] != 0 and 'output' in res:
                clear_output(output=res['output'])

        for proc_result in [x for x in self.proc_results.values() if x['ret_code'] != 0]:
            if 'stdout' in proc_result and proc_result['stdout']:
                sys.stdout.write(proc_result['stdout'])
            if 'stderr' in proc_result and proc_result['stderr']:
                sys.stderr.write(proc_result['stderr'])
            if 'exception' in proc_result:
                excp = proc_result['exception']
                if isinstance(excp, StopInputGroup):
                    if excp.message:
                        env.logger.info(excp.message)
                    self.output_groups[proc_result['index']] = sos_targets([])
                elif isinstance(excp, RemovedTarget):
                    raise excp
                elif 'task' in proc_result:
                    if env.config['error_mode'] == 'ignore':
                        env.logger.warning(
                            f"Ignore failed task {proc_result['task']}.")
                    # if the exception is from a task...
                    self.exec_error.append(proc_result['task'], excp)
            else:
                self.exec_error.append(
                    RuntimeError(
                        f"Substep failed with return code {proc_result['ret_code']}"
                    ))

        # this is after all substeps have been completed
        if self.exec_error.errors:
            raise self.exec_error

    def calculate_completed(self):
        substeps = self.completed['__substep_completed__'] + self.completed[
            '__substep_skipped__']
        self.completed['__step_completed__'] = self.completed[
            '__substep_completed__'] / substeps
        self.completed['__step_skipped__'] = self.completed[
            '__substep_skipped__'] / substeps
        if self.completed['__step_completed__'].is_integer():
            self.completed['__step_completed__'] = int(
                self.completed['__step_completed__'])
        if self.completed['__step_skipped__'].is_integer():
            self.completed['__step_skipped__'] = int(
                self.completed['__step_skipped__'])

    def run(self):
        '''Execute a single step and return results. The result for batch mode is the
        input, output etc returned as alias, and for interactive mode is the return value
        of the last expression. '''
        # return value of the last executed statement
        self.last_res = None
        self.start_time = time.time()
        self.completed = defaultdict(int)
        #
        # prepare environments, namely variables that can be used by the step
        #
        # * step_name:  name of the step, can be used by step process to determine
        #               actions dynamically.
        env.sos_dict.set('step_name', self.step.step_name())
        env.sos_dict.set('__last_step__', self.step.last_step)
        self.log('start')
        env.sos_dict.set(
            'step_id',
            textMD5(
                f'{env.sos_dict["workflow_id"]} {env.sos_dict["step_name"]} {self.step.md5}'
            ))
        env.sos_dict.set('master_id', env.config['master_id'])
        # used by nested workflow
        env.sos_dict.set('__step_context__', self.step.context)

        env.sos_dict.set('_runtime', {})
        # * input:      input files, which should be __step_output__ if it is defined, or
        #               None otherwise.
        # * _input:     first batch of input, which should be input if no input statement is used
        # * output:     None at first, can be redefined by output statement
        # * _output:    None at first, can be redefined by output statement
        # * depends:    None at first, can be redefined by depends statement
        # * _depends:   None at first, can be redefined by depends statement
        #
        self.init_input_output_vars()

        # _index is needed for pre-input action's active option and for debug output of scripts
        env.sos_dict.set('_index', 0)

        if 'STEP' in env.config['SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
            env.log_to_file(
                'STEP',
                f'Executing step {env.sos_dict["step_name"]} with step_input {env.sos_dict["step_input"]} and step_output {env.sos_dict["step_output"]}'
            )

        self.set_task_queue_from_task_params()

        # look for input statement.
        input_statement_idx = [
            idx for idx, x in enumerate(self.step.statements)
            if x[0] == ':' and x[1] == 'input'
        ]
        if not input_statement_idx:
            input_statement_idx = None
        elif len(input_statement_idx) == 1:
            input_statement_idx = input_statement_idx[0]
        else:
            raise ValueError(
                f'More than one step input are specified in step {self.step.step_name(True)}'
            )

        # if shared is true, we have to disable concurrent because we
        # do not yet return anything from shared.
        self.concurrent_substep = 'shared' not in self.step.options
        # and \
        #     ('concurrent' not in env.sos_dict['_runtime'] or env.sos_dict['_runtime']['concurrent'] is True)
        if input_statement_idx is not None:
            # execute before input stuff
            for statement in self.step.statements[:input_statement_idx]:
                if statement[0] == ':':
                    # wait for all dependent targets to be resolved to be resolved
                    key, value = statement[1:3]
                    if key != 'depends':
                        raise ValueError(
                            f'Step input should be specified before {key}')
                    while True:
                        try:
                            args, kwargs = SoS_eval(
                                f'__null_func__({value})',
                                extra_dict={
                                    '__null_func__': __null_func__,
                                    'output_from': __output_from__,
                                    'named_output': __named_output__,
                                    'traced': __traced__
                                })
                            dfiles = expand_depends_files(*args)
                            # dfiles can be Undetermined
                            runner = self.process_depends_args(dfiles, **kwargs)
                            try:
                                yreq = next(runner)
                                while True:
                                    yres = yield yreq
                                    yreq = runner.send(yres)
                            except StopIteration:
                                pass
                        except (UnknownTarget, RemovedTarget) as e:
                            runner = self.handle_unknown_target(e)
                            try:
                                yreq = next(runner)
                                while True:
                                    yres = yield yreq
                                    yreq = runner.send(yres)
                            except StopIteration:
                                pass
                            continue
                        except UnavailableLock:
                            raise
                        except Exception as e:
                            raise RuntimeError(
                                f'Failed to process step {key} ({value.strip()}): {e}'
                            )
                        break
                else:
                    try:
                        # 1354
                        # if there are definition before input, the definitions and imports
                        # must be added to global_def in order to be executed by substeps
                        if any(x in statement[1]
                               for x in ('class', 'def', 'import')):
                            step_def = KeepOnlyImportAndDefine().visit(
                                ast.parse(statement[1]))
                            if step_def.body:
                                if isinstance(self.step.global_def, ast.Module):
                                    self.step.global_def.body.extend(
                                        step_def.body)
                                else:
                                    self.step.global_def = step_def
                        self.execute(statement[1])
                    except StopInputGroup as e:
                        # stop before substeps, because there is no output statement before it
                        # we do not have to worry about keep_output
                        if e.message:
                            env.logger.info(e.message)
                        return self.collect_result()
            # input statement
            stmt = self.step.statements[input_statement_idx][2]
            self.log('input statement', stmt)
            while True:
                # wait for all targets to be resovled
                try:
                    args, kwargs = SoS_eval(
                        f"__null_func__({stmt})",
                        extra_dict={
                            '__null_func__': __null_func__,
                            'output_from': __output_from__,
                            'named_output': __named_output__,
                            'traced': __traced__
                        })
                    # Files will be expanded differently with different running modes
                    input_files: sos_targets = expand_input_files(
                        *args, **{
                            k: v
                            for k, v in kwargs.items()
                            if k not in SOS_INPUT_OPTIONS
                        })
                    runner = self.process_input_args(
                        input_files, **{
                            k: v
                            for k, v in kwargs.items()
                            if k in SOS_INPUT_OPTIONS
                        })
                    try:
                        yreq = next(runner)
                        while True:
                            yres = yield yreq
                            yreq = runner.send(yres)
                    except StopIteration as e:
                        self._substeps = e.value
                    #
                    if 'concurrent' in kwargs and self.concurrent_substep:
                        # concurrent can be True/False or an integer
                        self.concurrent_substep = kwargs['concurrent']
                except (UnknownTarget, RemovedTarget) as e:
                    runner = self.handle_unknown_target(e)
                    try:
                        yreq = next(runner)
                        while True:
                            yres = yield yreq
                            yreq = runner.send(yres)
                    except StopIteration:
                        pass
                    continue
                except UnavailableLock:
                    raise
                except Exception as e:
                    raise ValueError(
                        f'Failed to process input statement {stmt}: {e}')
                break

            input_statement_idx += 1
        elif env.sos_dict['step_input'].groups:
            # if default has groups...
            # default case
            self._substeps = env.sos_dict['step_input'].groups
            # assuming everything starts from 0 is after input
            input_statement_idx = 0
        else:
            # default case
            self._substeps = [env.sos_dict['step_input']]
            # assuming everything starts from 0 is after input
            input_statement_idx = 0

        self.proc_results = {}
        self.vars_to_be_shared = set()
        if 'shared' in self.step.options:
            self.vars_to_be_shared = parse_shared_vars(
                self.step.options['shared'])
        self.vars_to_be_shared = sorted([
            x[5:] if x.startswith('step_') else x
            for x in self.vars_to_be_shared
            if x not in ('step_', 'step_input', 'step_output', 'step_depends')
        ])
        self.shared_vars = [{} for x in self._substeps]
        # run steps after input statement, which will be run multiple times for each input
        # group.
        env.sos_dict.set('__num_groups__', len(self._substeps))

        # determine if a single index or the whole step should be skipped
        skip_index = False
        # signatures of each index, which can remain to be None if no output
        # is defined.
        self.output_groups = [sos_targets([]) for x in self._substeps]
        self.depends_groups = [sos_targets([]) for x in self._substeps]

        # used to prevent overlapping output from substeps
        self._all_outputs = set()
        self._subworkflow_results = []

        if any('sos_run' in x[1] for x in self.step.statements[input_statement_idx:]) and \
            'shared' not in self.step.options and not self.step.task and \
            self.step.statements[-1][0] == '!' and \
            (len(self.step.statements) == 1 or self.step.statements[-2][0] == ':') and \
            is_sos_run_the_only_last_stmt(self.step.statements[-1][1]):
            env.sos_dict.set('__concurrent_subworkflow__', True)

        if self.concurrent_substep:
            if len(self._substeps) <= 1 or env.config['run_mode'] == 'dryrun':
                self.concurrent_substep = False
            elif any('sos_run' in x[1]
                     for x in self.step.statements[input_statement_idx:]):
                self.concurrent_substep = False
                env.logger.debug(
                    'Substeps are executed sequentially because of existence of multiple nested workflow.'
                )
            else:
                self.prepare_substep()

        try:
            self.completed['__substep_skipped__'] = 0
            self.completed['__substep_completed__'] = len(self._substeps)
            self._completed_concurrent_substeps = 0
            # pending signatures are signatures for steps with external tasks
            self.pending_signatures = [None for x in self._substeps]

            for idx, g in enumerate(self._substeps):
                #
                # https://github.com/vatlab/sos/issues/1376
                #
                #   [default]
                #   input: for_each=dict(i=range(1000))
                #   sos_run('a', t=i)
                #
                # when we have workflow like the following when steps
                # are executed quickly with subworkflows submitted to the master
                # the master process could be swamped with subworkflows, causing
                # "too many open files".
                #
                # the following code will stop the step from continued
                # execution and wait for the subworkflows to complete.
                #
                if self._subworkflow_results:
                    try:
                        runner = self.wait_for_subworkflows(
                            allow_pending=env.config['worker_procs'])
                        yreq = next(runner)
                        while True:
                            yres = yield yreq
                            yreq = runner.send(yres)
                    except StopIteration:
                        pass

                # other variables
                #
                _vars = {}
                # now, let us expose target level variables as lists
                if len(g) > 1:
                    names = set.union(
                        *[set(x._dict.keys()) for x in g._targets])
                elif len(g) == 1:
                    names = set(g._targets[0]._dict.keys())
                else:
                    names = set()
                for name in names:
                    _vars[name] = [x.get(name) for x in g._targets]
                # then we expose all group level variables
                _vars.update(g._dict)
                _vars.update(env.sos_dict['step_input']._dict)
                env.sos_dict.update(_vars)

                env.sos_dict.set('_input', copy.deepcopy(g))
                # set vars to _input
                #env.sos_dict['_input'].set(**v)

                self.log('_input')
                env.sos_dict.set('_index', idx)

                if env.config['error_mode'] == 'ignore':
                    missed = [x for x in g.targets if not x.target_exists()]
                    if missed:
                        if any(isinstance(x, invalid_target) for x in missed):
                            env.logger.warning(
                                f'{self.step.step_name(True)}{f" (index={idx})" if len(self._substeps) > 1 else ""} ignored due to invalid input caused by previous failed substep.'
                            )
                        else:
                            env.logger.warning(
                                f'{self.step.step_name(True)}{f" (index={idx})" if len(self._substeps) > 1 else ""} ignored due to missing input {sos_targets(missed)}'
                            )
                        self.output_groups[idx] = sos_targets(invalid_target())
                        env.sos_dict.set('_output',
                                         sos_targets(invalid_target()))
                        self.skip_substep()
                        continue

                # in interactive mode, because sos_dict are always shared
                # execution of a substep, especially when it calls a nested
                # workflow, would change step_name, __step_context__ etc, and
                # we will have to reset these variables to make sure the next
                # substep would execute normally. Batch mode is immune to this
                # problem because nested workflows are executed in their own
                # process/context etc
                if env.config['run_mode'] == 'interactive':
                    env.sos_dict.set('step_name', self.step.step_name())
                    env.sos_dict.set(
                        'step_id',
                        hash((env.sos_dict["workflow_id"],
                              env.sos_dict["step_name"], self.step.md5)))
                    # used by nested workflow
                    env.sos_dict.set('__step_context__', self.step.context)
                #
                pre_statement = []
                if not any(st[0] == ':' and st[1] == 'output' for st in self.step.statements[input_statement_idx:]) and \
                        '__default_output__' in env.sos_dict:
                    pre_statement = [[':', 'output', '_output']]

                # if there is no statement, no task, claim success
                post_statement = []
                if not self.step.statements or self.step.statements[-1][
                        0] != '!':
                    if self.step.task:
                        # if there is only task, we insert a fake statement so that it can be executed by the executor
                        post_statement = [['!', '']]
                    else:
                        # complete case: no step, no statement
                        send_message_to_controller([
                            'progress', 'substep_completed',
                            env.sos_dict['step_id']
                        ])

                all_statements = pre_statement + self.step.statements[
                    input_statement_idx:] + post_statement
                self.is_input_verified = True
                for statement_idx, statement in enumerate(all_statements):
                    is_last_runblock = statement_idx == len(all_statements) - 1

                    # if input is undertermined, we can only process output:
                    if not g.valid() and statement[0] != ':':
                        raise RuntimeError('Undetermined input encountered')
                    if statement[0] == ':':
                        key, value = statement[1:3]
                        # output, depends, and process can be processed multiple times
                        while True:
                            # loop for all unresolved targets to be resolved
                            try:
                                args, kwargs = SoS_eval(
                                    f'__null_func__({value})',
                                    extra_dict={
                                        '__null_func__': __null_func__,
                                        'output_from': __output_from__,
                                        'named_output': __named_output__,
                                        'traced': __traced__
                                    })
                                # dynamic output or dependent files
                                if key == 'output':
                                    # if output is defined, its default value needs to be cleared
                                    if idx == 0:
                                        env.sos_dict.set(
                                            'step_output', sos_targets())
                                    ofiles: sos_targets = expand_output_files(
                                        value, *args, **{
                                            k: v
                                            for k, v in kwargs.items()
                                            if k not in SOS_OUTPUT_OPTIONS
                                        })
                                    if g.valid() and ofiles.valid():
                                        if any(
                                                x in g._targets
                                                for x in ofiles
                                                if not isinstance(x, sos_step)):
                                            raise RuntimeError(
                                                f'Overlapping input and output files: {", ".join(repr(x) for x in ofiles if x in g)}'
                                            )

                                    # set variable _output and output
                                    self.process_output_args(
                                        ofiles, **{
                                            k: v
                                            for k, v in kwargs.items()
                                            if k in SOS_OUTPUT_OPTIONS
                                        })
                                    self.output_groups[idx] = env.sos_dict[
                                        '_output']
                                elif key == 'depends':
                                    try:
                                        dfiles = expand_depends_files(*args)
                                        # dfiles can be Undetermined
                                        runner = self.process_depends_args(
                                            dfiles, **kwargs)
                                        try:
                                            yreq = next(runner)
                                            while True:
                                                yres = yield yreq
                                                yreq = runner.send(yres)
                                        except StopIteration:
                                            pass
                                        self.depends_groups[idx] = env.sos_dict[
                                            '_depends']
                                        self.log('_depends')
                                    except Exception:
                                        #env.logger.info(e)
                                        raise
                                else:
                                    raise RuntimeError(
                                        f'Unrecognized directive {key}')
                                # everything is ok, break
                                break
                            except (UnknownTarget, RemovedTarget) as e:
                                runner = self.handle_unknown_target(e)
                                try:
                                    yreq = next(runner)
                                    while True:
                                        yres = yield yreq
                                        yreq = runner.send(yres)
                                except StopIteration:
                                    pass
                                continue
                            except UnavailableLock:
                                raise
                            except Exception as e:
                                # if input is Undertermined, it is possible that output cannot be processed
                                # due to that, and we just return
                                if not g.valid():
                                    env.logger.debug(e)
                                    return self.collect_result()
                                raise RuntimeError(
                                    f'Failed to process step {key} ({value.strip()}): {e}'
                                )
                    elif is_last_runblock:
                        if env.config['sig_mode'] == 'skip' and not self.vars_to_be_shared and 'sos_run' not in statement[1] \
                            and not env.sos_dict['_output'].unspecified() and len(env.sos_dict['_output']) > 0 \
                            and all(x.target_exists() for x in env.sos_dict['_output'].targets) \
                            and env.sos_dict['_output'].later_than(env.sos_dict['_input']):
                            self.skip_substep()
                            env.logger.info(
                                f'``{env.sos_dict["step_name"]}``{f" (index={idx})" if len(self._substeps) > 1 else ""} is ``skipped`` with existing output.'
                            )
                            skip_index = True
                            # do not execute the rest of the statement
                            break
                        #
                        # default mode, check if skipping substep
                        sig = None
                        if env.config['sig_mode'] not in (
                                'ignore', 'distributed', 'build'
                        ) and not env.sos_dict['_output'].unspecified():
                            sig = RuntimeInfo(
                                statementMD5([statement[1], self.step.task]),
                                env.sos_dict['_input'],
                                env.sos_dict['_output'],
                                env.sos_dict['_depends'],
                                env.sos_dict['__signature_vars__'],
                                shared_vars=self.vars_to_be_shared)
                            matched = validate_step_sig(sig)
                            skip_index = bool(matched)
                            if skip_index:
                                # matched["output"] might hav vars not defined in "output" #1355
                                env.sos_dict.set('_output', matched["output"])
                                self.output_groups[idx] = matched["output"]
                                if 'vars' in matched:
                                    self.shared_vars[idx].update(
                                        matched["vars"])
                                self.skip_substep()

                                break
                        try:
                            if self.concurrent_substep:
                                self.concurrent_exec(statement, sig)
                                # we check if the previous task has been completed and process them
                                # because further steps might need to be done
                                try:
                                    runner = self.process_returned_substep_result(
                                        till=idx + 1, wait=False)
                                    yreq = next(runner)
                                    while True:
                                        yres = yield yreq
                                        yreq = runner.send(yres)
                                except StopIteration:
                                    pass
                            elif env.config[
                                    'sig_mode'] == 'ignore' or env.sos_dict[
                                        '_output'].unspecified():
                                self.local_exec_without_signature(statement)
                            else:
                                skip_index = self.local_exec_with_signature(
                                    statement, sig)
                                if skip_index:
                                    self.skip_substep()
                                    break

                        except StopInputGroup as e:
                            if not e.keep_output:
                                clear_output()
                                self.output_groups[idx] = sos_targets([])
                            if e.message:
                                env.logger.info(e.message)
                            skip_index = True
                            break
                        except Exception as e:
                            clear_output()
                            if env.config['error_mode'] == 'abort':
                                raise
                            elif env.config['error_mode'] == 'ignore':
                                idx_msg = f'(id={env.sos_dict["step_id"]}, index={idx})' if len(
                                    self._substeps
                                ) > 1 else f'(id={env.sos_dict["step_id"]})'
                                env.logger.warning(
                                    f'{self.step.step_name(True)} {idx_msg} returns no output due to error: {e}'
                                )
                                self.output_groups[idx] = sos_targets(
                                    invalid_target())
                                skip_index = True
                            else:
                                if env.config['run_mode'] != 'interactive':
                                    # default mode
                                    idx_msg = f'(id={env.sos_dict["step_id"]}, index={idx})' if len(
                                        self._substeps
                                    ) > 1 else f'(id={env.sos_dict["step_id"]})'
                                    env.logger.error(
                                        f'{self.step.step_name(True)} {idx_msg} returns an error.'
                                    )
                                self.exec_error.append(str(idx), e)
                    else:
                        # if it is not the last statement group (e.g. statements before :output)
                        # we execute locally without anything like signature
                        if self.is_input_verified:
                            verify_input()
                            self.is_input_verified = False
                        try:
                            self.execute(statement[1])
                        except StopInputGroup as e:
                            if not e.keep_output:
                                clear_output()
                                self.output_groups[idx] = sos_targets([])
                            if e.message:
                                env.logger.info(e.message)
                            skip_index = True
                            break
                        except Exception:
                            clear_output()
                            raise
                # if there is no statement , but there are tasks, we should
                # check signature here.
                if (not self.step.statements or self.step.statements[-1][0] != '!') and self.step.task and not self.concurrent_substep \
                    and env.config['sig_mode'] != 'ignore' and not env.sos_dict['_output'].unspecified():
                    skip_index = self.check_task_sig()

                # if this index is skipped, go directly to the next one
                if skip_index:
                    self.completed['__substep_skipped__'] += 1
                    self.completed['__substep_completed__'] -= 1
                    skip_index = False
                    continue

                # if concurrent input group, tasks are handled in substep
                if self.concurrent_substep or not self.step.task:
                    continue

                if env.config[
                        'run_mode'] == 'dryrun' and env.sos_dict['_index'] != 0:
                    continue

                # # check if the task is active
                # if 'active' in env.sos_dict['_runtime']:
                #     if not self.is_task_active():
                #         continue
                #
                self.log('task')
                try:
                    task_id, taskdef, task_vars = create_task(
                        self.step.global_def, self.step.global_vars,
                        self.step.task, self.step.task_params)
                    task = self.submit_task({
                        'index': env.sos_dict['_index'],
                        'task_id': task_id,
                        'task_def': taskdef,
                        'task_vars': task_vars
                    })
                    self.proc_results[env.sos_dict['_index']] = task
                except Exception as e:
                    # FIXME: cannot catch exception from subprocesses
                    if env.verbosity > 2:
                        sys.stderr.write(get_traceback())
                    raise RuntimeError(
                        f'Failed to execute process\n"{short_repr(self.step.task)}"\n{e}'
                    )

                #
                # # if not concurrent, we have to wait for the completion of the task
                # if 'concurrent' in env.sos_dict['_runtime'] and env.sos_dict[
                #         '_runtime']['concurrent'] is False:
                #     # in this case the steps must be executed not concurrently
                #     runner = self.wait_for_results(all_submitted=False)
                #     try:
                #         yreq = next(runner)
                #         while True:
                #             yres = yield yreq
                #             yreq = runner.send(yres)
                #     except StopIteration:
                #         pass
                #
                # endfor loop for each input group
                #
            if self._subworkflow_results:
                try:
                    runner = self.wait_for_subworkflows(allow_pending=0)
                    yreq = next(runner)
                    while True:
                        yres = yield yreq
                        yreq = runner.send(yres)
                except StopIteration:
                    pass
                env.sos_dict.pop('__concurrent_subworkflow__')

            runner = self.wait_for_results(all_submitted=True)
            try:
                yreq = next(runner)
                while True:
                    yres = yield yreq
                    yreq = runner.send(yres)
            except StopIteration:
                pass

            for idx, res in self.proc_results.items():
                if 'sig_skipped' in res:
                    self.completed['__substep_skipped__'] += 1
                    self.completed['__substep_completed__'] -= 1
                if 'output' in res:
                    self.output_groups[idx] = res["output"]

            # check results
            self.check_results()

            # if error happened but we allow all substeps to be completed, we now
            # raise exception
            if self.exec_error.errors:
                raise self.exec_error

            # if output is Undetermined, re-evalulate it
            # finalize output from output_groups because some output might be skipped
            # this is the final version of the output but we do maintain output
            # during the execution of step, for compatibility.
            env.sos_dict.set('step_output',
                             sos_targets([])._add_groups(self.output_groups))
            env.sos_dict.set('step_depends',
                             sos_targets([])._add_groups(self.depends_groups))

            # if there exists an option shared, the variable would be treated as
            # provides=sos_variable(), and then as step_output
            if 'shared' in self.step.options:
                self.shared_vars = evaluate_shared(self.shared_vars,
                                                   self.step.options['shared'])
                env.sos_dict.quick_update(self.shared_vars)
            missing = self.verify_output()
            self.log(
                'output',
                msg=f'\033[95m missing: {short_repr(missing)} ({len(missing)} item{"s" if len(missing)>1 else ""})\033[0m'
                if len(missing) > 0 else '')
            self.calculate_completed()

            def file_only(targets):
                if not isinstance(targets, sos_targets):
                    env.logger.warning(
                        f"Unexpected input or output target for reporting. Empty list returned: {targets}"
                    )
                    return []
                return [(str(x), x.size())
                        for x in targets._targets
                        if isinstance(x, file_target)]

            step_info = {
                'step_id': self.step.md5,
                'start_time': self.start_time,
                'stepname': self.step.step_name(True),
                'substeps': len(self._substeps),
                'input': file_only(env.sos_dict['step_input']),
                'output': file_only(env.sos_dict['step_output']),
                'completed': dict(self.completed),
                'end_time': time.time()
            }
            send_message_to_controller([
                'workflow_sig', 'step', env.sos_dict["workflow_id"],
                repr(step_info)
            ])
            return self.collect_result()
        finally:
            if self.concurrent_substep:
                close_socket(self.result_pull_socket, 'substep collector')


class Step_Executor(Base_Step_Executor):
    '''Single process step executor'''

    def __init__(self, step, socket, mode='run'):
        self.run_mode = mode
        env.config['run_mode'] = mode
        super(Step_Executor, self).__init__(step)
        self.socket = socket
        # because step is executed in a separate SoS_Worker process, this
        # __socket__ is available to all the actions that will be executed
        # in the step
        env.__socket__ = socket

    def submit_tasks(self, tasks):
        if 'TASK' in env.config['SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
            env.log_to_file('TASK', f'Send {tasks}')
        self.socket.send(
            encode_msg(['tasks', env.sos_dict['_runtime']['queue']] + tasks))

    def wait_for_tasks(self, tasks, all_submitted):
        # wait for task is a generator function that yields the request
        # to the runner
        if not tasks:
            return {}
        # when we wait, the "outsiders" also need to see the tags etc
        # of the tasks so we have to write to the database. #156
        send_message_to_controller(['commit_sig'])

        # wait till the executor responde
        results = {}
        while True:
            # yield an indicator of what is requested, for debugging purpose
            yield self.socket
            res = decode_msg(self.socket.recv())
            if res is None:
                sys.exit(0)
            results.update(res)

            # all results have been obtained.
            if len(results) == len(tasks):
                break
        return results

    def wait_for_subworkflows(self, allow_pending):
        '''Wait for results from subworkflows'''
        try:
            allow_pending = int(allow_pending)
        except:
            allow_pending = min(max(os.cpu_count() // 2, 2), 8)

        while self._subworkflow_results:
            if allow_pending > 0:
                n_pending = sum(
                    len(x['pending_workflows'])
                    for x in self._subworkflow_results)
                if n_pending <= allow_pending:
                    break
            # here we did not check if workflow ids match
            yield self.socket
            res = decode_msg(self.socket.recv())
            if res is None:
                sys.exit(0)
            elif isinstance(res, Exception):
                raise res
            if not '__workflow_id__' in res:
                raise ValueError(
                    f'Unrecognized result from subworkflows: {res}')
            # remove from _self._subworkflow_results
            result_with_id = [
                idx for idx, x in enumerate(self._subworkflow_results)
                if res['__workflow_id__'] in x['pending_workflows']
            ]
            if not result_with_id:
                raise RuntimeError(
                    f'Failed to identify ID of returned subworkflow: {res}')
            if len(result_with_id) > 1:
                raise RuntimeError(
                    'Multiple matches of subworkflow ID. This should not happen.'
                )
            self._subworkflow_results[
                result_with_id[0]]['pending_workflows'].remove(
                    res['__workflow_id__'])
            if not self._subworkflow_results[
                    result_with_id[0]]['pending_workflows']:
                self._subworkflow_results.pop(result_with_id[0])

    def handle_unknown_target(self, e):
        self.socket.send(encode_msg(['missing_target', e.target]))
        yield self.socket
        res = decode_msg(self.socket.recv())
        if not res:
            raise e

    def verify_dynamic_targets(self, targets):
        if not targets:
            return

        if env.config['trace_existing']:
            traced = targets
        else:
            traced = [x for x in targets if x.traced]

        if not traced:
            return

        self.socket.send(encode_msg(['dependent_target'] + traced))
        yield self.socket
        res = decode_msg(self.socket.recv())
        if res != 'target_resolved':
            raise RuntimeError(f'Failed to veryify dependent target {traced}')

    def run(self):
        try:
            try:
                # 1218
                runner = Base_Step_Executor.run(self)
                yreq = next(runner)
                while True:
                    yres = yield yreq
                    yreq = runner.send(yres)
            except StopIteration as e:
                res = e.value

            if self.socket is not None:
                if 'STEP' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                        'SOS_DEBUG']:
                    env.log_to_file(
                        'STEP',
                        f'Step {self.step.step_name()} sends result {short_repr(res)}'
                    )
                self.socket.send(encode_msg(res))
            else:
                return res
        except RemovedTarget as e:
            # removed target needs to be handled differently since the workflow manager
            # use type information to get removed targets
            if self.socket is not None and not self.socket.closed:
                self.socket.send(encode_msg(e))
            else:
                raise e
        except Exception as e:
            if env.verbosity > 2:
                sys.stderr.write(get_traceback())
            if isinstance(e, ProcessKilled):
                raise
            # if not self.exec_error
            if e != self.exec_error:
                self.exec_error.append(self.step.step_name(), e)
        #
        if self.exec_error.errors:
            if self.socket is not None and not self.socket.closed:
                env.log_to_file(
                    'STEP',
                    f'Step {self.step.step_name()} sends exception {self.exec_error}'
                )
                self.socket.send(encode_msg(self.exec_error))
            else:
                raise self.exec_error



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import copy
import os
import subprocess

from .eval import cfg_interpolate
from .utils import env
from .targets import textMD5


class WorkflowEngine:

    def __init__(self, agent):
        self.agent = agent
        self.config = agent.config
        self.alias = self.config['alias']

    def remove_arg(self, argv, arg):
        r_idx = [idx for idx, x in enumerate(argv) if x.startswith(arg)]
        if not r_idx:
            return argv
        else:
            r_idx = r_idx[0]
        # find next option
        r_next = [
            idx for idx, x in enumerate(argv[r_idx + 1:]) if x.startswith('-')
        ]
        if r_next:
            argv = argv[:r_idx] + argv[r_idx + 1 + r_next[0]:]
        else:
            argv = argv[:r_idx]
        return argv

    def expand_template(self):
        try:
            if self.local_filename.lower().endswith('.ipynb'):
                from .converter import extract_workflow
                script = extract_workflow(self.local_filename)
            else:
                with open(self.local_filename) as script_file:
                    script = script_file.read()
            self.job_name = textMD5(script)
            self.template_args['filename'] = self.filename
            self.template_args['script'] = script
            self.template_args['command'] = self.command
            self.template_args['job_name'] = self.job_name
            self.job_text = cfg_interpolate(self.workflow_template,
                                            self.template_args) + '\n'
        except Exception as e:
            raise ValueError(
                f'Failed to generate job file for the execution of workflow: {e}'
            )
        try:
            wf_dir = os.path.join(os.path.expanduser('~'), '.sos', 'workflows')
            if not os.path.isdir(wf_dir):
                os.makedirs(wf_dir)

            self.job_file = os.path.join(wf_dir, self.job_name + '.sh')

            # do not translate newline under windows because the script will be executed
            # under linux/mac
            with open(self.job_file, 'w', newline='') as job:
                job.write(self.job_text)
        except Exception as e:
            raise RuntimeError(
                f'Failed to submit workflow {self.command} with script \n{self.job_text}\n: {e}'
            )
        return True

    def execute_workflow(self, filename, command, **template_args):
        # there is no need to prepare workflow (e.g. copy over)
        if os.path.isfile(filename):
            ret = self.agent.send_to_host([filename])
        elif os.path.isfile(filename + '.sos'):
            ret = self.agent.send_to_host([filename + '.sos'])
        elif os.path.isfile(filename + '.ipynb'):
            ret = self.agent.send_to_host([filename + '.ipynb'])
        else:
            raise RuntimeError(f'Failed to locate script {filename}')

        self.local_filename = filename
        self.filename = list(ret.values())[0]
        self.command = self.remove_arg(command, '-r')
        # -c only point to local config file.
        self.command = self.remove_arg(self.command, '-c')
        # remove --slave mode because the master cannot reach remote slave
        self.command = self.remove_arg(self.command, '-m')
        # replace absolute path with relative one because remote sos might have
        # a different path.
        if os.path.basename(command[0]) == 'sos':
            self.command[0] = 'sos'
            self.command[2] = self.filename
        elif os.path.basename(command[0]) == 'sos-runner':
            self.command[0] = 'sos-runner'
            self.command[1] = self.filename
        else:
            raise ValueError(
                f'Failed to generate remote execution command: {self.command}')
        self.command = subprocess.list2cmdline(self.command)
        self.template_args = copy.deepcopy(self.config)
        self.template_args.update(template_args)
        env.log_to_file('WORKFLOW',
                        f'Execute command on remote host: {self.command}')
        return True


class BackgroundProcess_WorkflowEngine(WorkflowEngine):

    def __init__(self, agent):
        super(BackgroundProcess_WorkflowEngine, self).__init__(agent)
        if 'workflow_template' in self.config:
            self.workflow_template = self.config['workflow_template'].replace(
                '\r\n', '\n')
        else:
            self.workflow_template = None

    def execute_workflow(self, filename, command, **template_args):
        #
        # calling super execute_workflow would set cleaned versions
        # of self.filename, self.command, and self.template_args
        if not super(BackgroundProcess_WorkflowEngine, self).execute_workflow(
                filename, command, **template_args):
            env.log_to_file(
                'WORKFLOW',
                f'Failed to prepare workflow with command "{command}"')
            return False

        if self.workflow_template:
            if not self._execute_workflow_with_template():
                return False
        else:
            if not self._execute_workflow():
                return False
        return True

    def _execute_workflow(self):
        # if no template, use a default command
        env.log_to_file('WORKDLOW', f'Execute "{self.command}"')
        try:
            self.agent.check_call(self.command, under_workdir=True)
        except Exception as e:
            raise RuntimeError(f'Failed to submit workflow {self.command}: {e}')
        return True

    def _execute_workflow_with_template(self):
        '''Submit workflows by interpolating a shell script defined in workflow_template'''
        self.expand_template()

        try:
            # then copy the job file to remote host if necessary
            self.agent.send_job_file(self.job_file, dir='workflows')

            cmd = f'bash ~/.sos/workflows/{os.path.basename(self.job_file)}'
            env.log_to_file(
                'WORKFLOW',
                f'Execute "{self.command}" with script {self.job_text}')
            self.agent.check_call(cmd, under_workdir=True)
        except Exception as e:
            raise RuntimeError(
                f'Failed to submit workflow {self.command} with script \n{self.job_text}\n: {e}'
            )
        finally:
            try:
                os.remove(self.job_file)
            except Exception as e:
                env.logger.debug(
                    f'Failed to remove temporary workflow file: {e}')
        return True



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import textwrap

from ._version import __version__

from .__main__ import get_run_parser
from .workflow_executor import Base_Executor

from sos.parser import SoS_Script

assert __version__


def execute_workflow(script: str,
                     workflow=None,
                     targets=None,
                     args=[],
                     options={},
                     config={}):
    '''
    Execute a SoS workflow with the following parameters:

    script:
        A (multi-line) string that defines one or more workflows.

    workflow (string, optional):
        Name of workflow to execute. This option can be ignored if
        ignored if the script defines a default workflow (with no name
        or with name `default`) or defines only a single workflow, or
        if the workflow is triggered by option "targets".

    targets (string or list, or sos_targets, optional):
        One (string) or more (list) of filenames as targets to be generated
        by the workflow, equivalent to option "-t" from command line.

    args (list or dict, optional):
        Command line arguments as a list in the format of "[`--cutoff', '0.5']"
        or a dictionary in the format of {"cutoff": 0.5} for workflow parameters
        defined in "parameter" statements. SoS options such as '-c', '-j',
        and '-s' should be defined in options.

    options (dict, optional):
        Dictionary with the following configuration options. Please
        refer to output of "sos run -h" for details about each option.
            config_file: configuration file ("-c"). The content of the config
                file can also be specified with option config.
            output_dag: option "-d"
            output_report: option "-p"
            default_queue: option "-q"
            worker_procs: option "-j"
            max_running_jobs: option "-J"
            sig_mode: option "-s"
            run_mode: "run" or "dryrun"
            verbosity: option "-v"
            trace_existing: option "-T"
            config_file: option "-c"

    config (dict, optional):
        config as if loaded from "options['config_file']"

    Note: executing on specified host (option "-r") is not supported by this function.
    '''
    try:
        script = SoS_Script(textwrap.dedent(script))
    except Exception as e:
        # show script with error
        raise ValueError(f'Failed to parse script {script}: {e}')
    #
    if workflow and targets:
        raise ValueError(
            "Only one of parameters workflow and targets should be specified.")
    wf = script.workflow(workflow, use_default=not targets)

    if not isinstance(config, dict):
        raise ValueError(f'Option config should be a dictionary.')
    run_options = {
        'config_file': None,
        'extra_config': config,
        'output_dag': None,
        'output_report': None,
        'default_queue': None,
        'worker_procs': None,
        'max_running_jobs': None,
        'sig_mode': 'default',
        'run_mode': 'run',
        'verbosity': 2,
        'workdir': os.getcwd(),
        'script': script,
        'workflow': workflow,
        'targets': targets,
        'workflow_args': args,
        'trace_existing': False,
        'error_mode': 'default',
        'exec_mode': None
    }
    # a convenience feature
    if isinstance(args, dict):
        run_options['workflow_vars'] = args
        workflow_args = []
    elif args:
        # for convenience,
        parser = get_run_parser(interactive=True, with_workflow=False)
        for arg in args:
            if arg.startswith(
                    '-') and not arg.startswith('--') and not arg in ['-c']:
                raise ValueError(
                    'SoS options should be specified with parameter "option"')
        # check args
        sos_args, workflow_args = parser.parse_known_args(args)
        if sos_args.__config__ and not 'config_file' in options:
            options['config_file'] = sos_args.__config__
    else:
        workflow_args = []

    run_options.update(options)

    from .utils import env
    env.verbosity = run_options['verbosity']

    executor = Base_Executor(wf, args=workflow_args, config=run_options)
    if isinstance(targets, str):
        targets = [targets]
    return executor.run(targets=targets)
